# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Sequence

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from monai.networks.blocks import Convolution
from monai.networks.blocks.spade_norm import SPADE
from monai.networks.layers import Act
from monai.networks.layers.utils import get_act_layer
from monai.utils.enums import StrEnum

__all__ = ["SPADENet"]


class UpsamplingModes(StrEnum):
    bicubic = "bicubic"
    nearest = "nearest"
    bilinear = "bilinear"


class SPADENetResBlock(nn.Module):
    """
    Creates a Residual Block with SPADE normalisation.

    Args:
        spatial_dims: number of spatial dimensions
        in_channels: number of input channels
        out_channels: number of output channels
        label_nc: number of semantic channels that will be taken into account in SPADE normalisation blocks
        spade_intermediate_channels: number of intermediate channels in the middle conv. layers in SPADE normalisation blocks
        norm: base normalisation type used on top of SPADE
        kernel_size: convolutional kernel size
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        label_nc: int,
        spade_intermediate_channels: int = 128,
        norm: str | tuple = "INSTANCE",
        act: str | tuple = (Act.LEAKYRELU, {"negative_slope": 0.2}),
        kernel_size: int = 3,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.int_channels = min(self.in_channels, self.out_channels)
        self.learned_shortcut = self.in_channels != self.out_channels
        self.conv_0 = Convolution(
            spatial_dims=spatial_dims, in_channels=self.in_channels, out_channels=self.int_channels, act=None, norm=None
        )
        self.conv_1 = Convolution(
            spatial_dims=spatial_dims,
            in_channels=self.int_channels,
            out_channels=self.out_channels,
            act=None,
            norm=None,
        )
        self.activation = get_act_layer(act)
        self.norm_0 = SPADE(
            label_nc=label_nc,
            norm_nc=self.in_channels,
            kernel_size=kernel_size,
            spatial_dims=spatial_dims,
            hidden_channels=spade_intermediate_channels,
            norm=norm,
        )
        self.norm_1 = SPADE(
            label_nc=label_nc,
            norm_nc=self.int_channels,
            kernel_size=kernel_size,
            spatial_dims=spatial_dims,
            hidden_channels=spade_intermediate_channels,
            norm=norm,
        )

        if self.learned_shortcut:
            self.conv_s = Convolution(
                spatial_dims=spatial_dims,
                in_channels=self.in_channels,
                out_channels=self.out_channels,
                act=None,
                norm=None,
                kernel_size=1,
            )
            self.norm_s = SPADE(
                label_nc=label_nc,
                norm_nc=self.in_channels,
                kernel_size=kernel_size,
                spatial_dims=spatial_dims,
                hidden_channels=spade_intermediate_channels,
                norm=norm,
            )

    def forward(self, x, seg):
        x_s = self.shortcut(x, seg)
        dx = self.conv_0(self.activation(self.norm_0(x, seg)))
        dx = self.conv_1(self.activation(self.norm_1(dx, seg)))
        out = x_s + dx
        return out

    def shortcut(self, x, seg):
        if self.learned_shortcut:
            x_s = self.conv_s(self.norm_s(x, seg))
        else:
            x_s = x
        return x_s


class SPADEEncoder(nn.Module):
    """
    Encoding branch of a VAE compatible with a SPADE-like generator

    Args:
        spatial_dims: number of spatial dimensions
        in_channels: number of input channels
        z_dim: latent space dimension of the VAE containing the image sytle information
        channels: number of output after each downsampling block
        input_shape: spatial input shape of the tensor, necessary to do the reshaping after the linear layers
        of the autoencoder (HxWx[D])
        kernel_size: convolutional kernel size
        norm: normalisation layer type
        act: activation type
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        z_dim: int,
        channels: Sequence[int],
        input_shape: Sequence[int],
        kernel_size: int = 3,
        norm: str | tuple = "INSTANCE",
        act: str | tuple = (Act.LEAKYRELU, {"negative_slope": 0.2}),
    ):
        super().__init__()
        self.in_channels = in_channels
        self.z_dim = z_dim
        self.channels = channels
        if len(input_shape) != spatial_dims:
            raise ValueError("Length of parameter input shape must match spatial_dims; got %s" % (input_shape))
        for s_ind, s_ in enumerate(input_shape):
            if s_ / (2 ** len(channels)) != s_ // (2 ** len(channels)):
                raise ValueError(
                    "Each dimension of your input must be divisible by 2 ** (autoencoder depth)."
                    "The shape in position %d, %d is not divisible by %d. " % (s_ind, s_, len(channels))
                )
        self.input_shape = input_shape
        self.latent_spatial_shape = [s_ // (2 ** len(self.channels)) for s_ in self.input_shape]
        blocks = []
        ch_init = self.in_channels
        for _, ch_value in enumerate(channels):
            blocks.append(
                Convolution(
                    spatial_dims=spatial_dims,
                    in_channels=ch_init,
                    out_channels=ch_value,
                    strides=2,
                    kernel_size=kernel_size,
                    norm=norm,
                    act=act,
                )
            )
            ch_init = ch_value

        self.blocks = nn.ModuleList(blocks)
        self.fc_mu = nn.Linear(
            in_features=np.prod(self.latent_spatial_shape) * self.channels[-1], out_features=self.z_dim
        )
        self.fc_var = nn.Linear(
            in_features=np.prod(self.latent_spatial_shape) * self.channels[-1], out_features=self.z_dim
        )

    def forward(self, x):
        for block in self.blocks:
            x = block(x)
        x = x.view(x.size(0), -1)
        mu = self.fc_mu(x)
        logvar = self.fc_var(x)
        return mu, logvar

    def encode(self, x):
        for block in self.blocks:
            x = block(x)
        x = x.view(x.size(0), -1)
        mu = self.fc_mu(x)
        logvar = self.fc_var(x)
        return self.reparameterize(mu, logvar)

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return eps.mul(std) + mu


class SPADEDecoder(nn.Module):
    """
    Decoder branch of a SPADE-like generator. It can be used independently, without an encoding branch,
    behaving like a GAN, or coupled to a SPADE encoder.

    Args:
        label_nc: number of semantic labels
        spatial_dims: number of spatial dimensions
        out_channels: number of output channels
        label_nc: number of semantic channels used for the SPADE normalisation blocks
        input_shape: spatial input shape of the tensor, necessary to do the reshaping after the linear layers
        channels: number of output after each downsampling block
        z_dim: latent space dimension of the VAE containing the image sytle information (None if encoder is not used)
        is_vae: whether the decoder is going to be coupled to an autoencoder or not (true: yes, false: no)
        spade_intermediate_channels: number of channels in the intermediate layers of the SPADE normalisation blocks
        norm: base normalisation type
        act:  activation layer type
        last_act: activation layer type for the last layer of the network (can differ from previous)
        kernel_size: convolutional kernel size
        upsampling_mode: upsampling mode (nearest, bilinear etc.)
    """

    def __init__(
        self,
        spatial_dims: int,
        out_channels: int,
        label_nc: int,
        input_shape: Sequence[int],
        channels: list[int],
        z_dim: int | None = None,
        is_vae: bool = True,
        spade_intermediate_channels: int = 128,
        norm: str | tuple = "INSTANCE",
        act: str | tuple = (Act.LEAKYRELU, {"negative_slope": 0.2}),
        last_act: str | tuple | None = (Act.LEAKYRELU, {"negative_slope": 0.2}),
        kernel_size: int = 3,
        upsampling_mode: str = UpsamplingModes.nearest.value,
    ):
        super().__init__()
        self.is_vae = is_vae
        self.out_channels = out_channels
        self.label_nc = label_nc
        self.num_channels = channels
        if len(input_shape) != spatial_dims:
            raise ValueError("Length of parameter input shape must match spatial_dims; got %s" % (input_shape))
        for s_ind, s_ in enumerate(input_shape):
            if s_ / (2 ** len(channels)) != s_ // (2 ** len(channels)):
                raise ValueError(
                    "Each dimension of your input must be divisible by 2 ** (autoencoder depth)."
                    "The shape in position %d, %d is not divisible by %d. " % (s_ind, s_, len(channels))
                )
        self.latent_spatial_shape = [s_ // (2 ** len(self.num_channels)) for s_ in input_shape]

        if not self.is_vae:
            self.conv_init = Convolution(
                spatial_dims=spatial_dims, in_channels=label_nc, out_channels=channels[0], kernel_size=kernel_size
            )
        elif self.is_vae and z_dim is None:
            raise ValueError(
                "If the network is used in VAE-GAN mode, parameter z_dim "
                "(number of latent channels in the VAE) must be populated."
            )
        else:
            self.fc = nn.Linear(z_dim, np.prod(self.latent_spatial_shape) * channels[0])

        self.z_dim = z_dim
        blocks = []
        channels.append(self.out_channels)
        self.upsampling = torch.nn.Upsample(scale_factor=2, mode=upsampling_mode)
        for ch_ind, ch_value in enumerate(channels[:-1]):
            blocks.append(
                SPADENetResBlock(
                    spatial_dims=spatial_dims,
                    in_channels=ch_value,
                    out_channels=channels[ch_ind + 1],
                    label_nc=label_nc,
                    spade_intermediate_channels=spade_intermediate_channels,
                    norm=norm,
                    kernel_size=kernel_size,
                    act=act,
                )
            )

        self.blocks = torch.nn.ModuleList(blocks)
        self.last_conv = Convolution(
            spatial_dims=spatial_dims,
            in_channels=channels[-1],
            out_channels=out_channels,
            padding=(kernel_size - 1) // 2,
            kernel_size=kernel_size,
            norm=None,
            act=last_act,
        )

    def forward(self, seg, z: torch.Tensor | None = None):
        """
        Args:
            seg: input BxCxHxW[xD] semantic map on which the output is conditioned on
            z: latent vector output by the encoder if self.is_vae is True. When is_vae is
            False, z is a random noise vector.

        Returns:

        """
        if not self.is_vae:
            x = F.interpolate(seg, size=tuple(self.latent_spatial_shape))
            x = self.conv_init(x)
        else:
            if (
                z is None and self.z_dim is not None
            ):  # Even though this network is a VAE (self.is_vae), you should be able to sample from noise as well.
                z = torch.randn(seg.size(0), self.z_dim, dtype=torch.float32, device=seg.get_device())
            x = self.fc(z)
            x = x.view(*[-1, self.num_channels[0]] + self.latent_spatial_shape)

        for res_block in self.blocks:
            x = res_block(x, seg)
            x = self.upsampling(x)

        x = self.last_conv(x)
        return x


class SPADENet(nn.Module):
    """
    SPADE Network, implemented based on the code by Park, T et al. in
    "Semantic Image Synthesis with Spatially-Adaptive Normalization"
    (https://github.com/NVlabs/SPADE)

    Args:
        spatial_dims: number of spatial dimensions
        in_channels: number of input channels
        out_channels: number of output channels
        label_nc: number of semantic channels used for the SPADE normalisation blocks
        input_shape:  spatial input shape of the tensor, necessary to do the reshaping after the linear layers
        channels: number of output after each downsampling block
        z_dim: latent space dimension of the VAE containing the image sytle information (None if encoder is not used)
        is_vae: whether the decoder is going to be coupled to an autoencoder (true) or not (false)
        spade_intermediate_channels: number of channels in the intermediate layers of the SPADE normalisation blocks
        norm: base normalisation type
        act: activation layer type
        last_act: activation layer type for the last layer of the network (can differ from previous)
        kernel_size: convolutional kernel size
        upsampling_mode: upsampling mode (nearest, bilinear etc.)
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        label_nc: int,
        input_shape: Sequence[int],
        channels: list[int],
        z_dim: int | None = None,
        is_vae: bool = True,
        spade_intermediate_channels: int = 128,
        norm: str | tuple = "INSTANCE",
        act: str | tuple = (Act.LEAKYRELU, {"negative_slope": 0.2}),
        last_act: str | tuple | None = (Act.LEAKYRELU, {"negative_slope": 0.2}),
        kernel_size: int = 3,
        upsampling_mode: str = UpsamplingModes.nearest.value,
    ):
        super().__init__()
        self.is_vae = is_vae
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.channels = channels
        self.label_nc = label_nc
        self.input_shape = input_shape

        if self.is_vae:
            if z_dim is None:
                ValueError("The latent space dimension mapped by parameter z_dim cannot be None is is_vae is True.")
            else:
                self.encoder = SPADEEncoder(
                    spatial_dims=spatial_dims,
                    in_channels=in_channels,
                    z_dim=z_dim,
                    channels=channels,
                    input_shape=input_shape,
                    kernel_size=kernel_size,
                    norm=norm,
                    act=act,
                )

        decoder_channels = channels
        decoder_channels.reverse()

        self.decoder = SPADEDecoder(
            spatial_dims=spatial_dims,
            out_channels=out_channels,
            label_nc=label_nc,
            input_shape=input_shape,
            channels=decoder_channels,
            z_dim=z_dim,
            is_vae=is_vae,
            spade_intermediate_channels=spade_intermediate_channels,
            norm=norm,
            act=act,
            last_act=last_act,
            kernel_size=kernel_size,
            upsampling_mode=upsampling_mode,
        )

    def forward(self, seg: torch.Tensor, x: torch.Tensor | None = None):
        z = None
        if self.is_vae:
            z_mu, z_logvar = self.encoder(x)
            z = self.encoder.reparameterize(z_mu, z_logvar)
            return self.decoder(seg, z), z_mu, z_logvar
        else:
            return (self.decoder(seg, z),)

    def encode(self, x: torch.Tensor):
        if self.is_vae:
            return self.encoder.encode(x)
        else:
            return None

    def decode(self, seg: torch.Tensor, z: torch.Tensor | None = None):
        return self.decoder(seg, z)



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.transforms import RandAffine
from monai.utils.type_conversion import convert_data_type
from tests.lazy_transforms_utils import test_resampler_lazy
from tests.utils import TEST_NDARRAYS_ALL, assert_allclose, is_tf32_env

_rtol = 1e-3 if is_tf32_env() else 1e-4

TESTS = []
for p in TEST_NDARRAYS_ALL:
    for device in [None, "cpu", "cuda"] if torch.cuda.is_available() else [None, "cpu"]:
        TESTS.append(
            [dict(device=device), {"img": p(torch.arange(27).reshape((3, 3, 3)))}, p(np.arange(27).reshape((3, 3, 3)))]
        )
        TESTS.append(
            [
                dict(device=device, spatial_size=-1),
                {"img": p(torch.arange(27).reshape((3, 3, 3)))},
                p(np.arange(27).reshape((3, 3, 3))),
            ]
        )
        TESTS.append(
            [
                dict(device=device),
                {"img": p(torch.arange(27).reshape((3, 3, 3))), "spatial_size": (2, 2)},
                p(np.array([[[2.0, 3.0], [5.0, 6.0]], [[11.0, 12.0], [14.0, 15.0]], [[20.0, 21.0], [23.0, 24.0]]])),
            ]
        )
        TESTS.append(
            [
                dict(device=device),
                {"img": p(torch.ones((1, 3, 3, 3))), "spatial_size": (2, 2, 2)},
                p(torch.ones((1, 2, 2, 2))),
            ]
        )
        TESTS.append(
            [
                dict(device=device, spatial_size=(2, 2, 2), cache_grid=True),
                {"img": p(torch.ones((1, 3, 3, 3)))},
                p(torch.ones((1, 2, 2, 2))),
            ]
        )
        TESTS.append(
            [
                dict(
                    prob=0.9,
                    rotate_range=(np.pi / 2,),
                    shear_range=[1, 2],
                    translate_range=[2, 1],
                    padding_mode="zeros",
                    spatial_size=(2, 2, 2),
                    device=device,
                ),
                {"img": p(torch.ones((1, 3, 3, 3))), "mode": "bilinear"},
                p(torch.tensor([[[[0.3658, 1.0000], [1.0000, 1.0000]], [[1.0000, 1.0000], [1.0000, 0.9333]]]])),
            ]
        )
        TESTS.append(
            [
                dict(
                    prob=0.9,
                    rotate_range=(np.pi / 2,),
                    shear_range=[1, 2],
                    translate_range=[2, 1],
                    padding_mode="zeros",
                    spatial_size=(2, 2, 2),
                    cache_grid=True,
                    device=device,
                ),
                {"img": p(torch.ones((1, 3, 3, 3))), "mode": "bilinear"},
                p(torch.tensor([[[[0.3658, 1.0000], [1.0000, 1.0000]], [[1.0000, 1.0000], [1.0000, 0.9333]]]])),
            ]
        )
        TESTS.append(
            [
                dict(
                    prob=0.9,
                    rotate_range=(np.pi / 2,),
                    shear_range=[1, 2],
                    translate_range=[2, 1],
                    scale_range=[0.1, 0.2],
                    device=device,
                ),
                {"img": p(torch.arange(64).reshape((1, 8, 8))), "spatial_size": (3, 3)},
                p(
                    torch.tensor(
                        [[[18.7362, 15.5820, 12.4278], [27.3988, 24.2446, 21.0904], [36.0614, 32.9072, 29.7530]]]
                    )
                ),
            ]
        )
        TESTS.append(
            [
                dict(
                    prob=0.9,
                    rotate_range=(np.pi / 2,),
                    shear_range=[1, 2],
                    translate_range=[2, 1],
                    scale_range=[0.1, 0.2],
                    spatial_size=(3, 3),
                    cache_grid=True,
                    device=device,
                ),
                {"img": p(torch.arange(64).reshape((1, 8, 8)))},
                p(
                    torch.tensor(
                        [[[18.7362, 15.5820, 12.4278], [27.3988, 24.2446, 21.0904], [36.0614, 32.9072, 29.7530]]]
                    )
                ),
            ]
        )

TEST_CASES_SKIPPED_CONSISTENCY = []
for p in TEST_NDARRAYS_ALL:
    for in_dtype in (np.int32, np.float32):
        TEST_CASES_SKIPPED_CONSISTENCY.append((p(np.arange(9 * 10).reshape(1, 9, 10)), in_dtype))

TEST_RANDOMIZE = []
for cache_grid in (False, True):
    for initial_randomize in (False, True):
        TEST_RANDOMIZE.append((initial_randomize, cache_grid))


class TestRandAffine(unittest.TestCase):

    @parameterized.expand(TESTS)
    def test_rand_affine(self, input_param, input_data, expected_val):
        g = RandAffine(**input_param)
        g.set_random_state(123)
        result = g(**input_data)
        g.rand_affine_grid.affine = torch.eye(4, dtype=torch.float64)  # reset affine
        test_resampler_lazy(g, result, input_param, input_data, seed=123, rtol=_rtol)
        if input_param.get("cache_grid", False):
            self.assertTrue(g._cached_grid is not None)
        assert_allclose(result, expected_val, rtol=_rtol, atol=1e-4, type_test="tensor")

    @parameterized.expand([(None,), ((1, 1, -1),)])
    def test_ill_cache(self, spatial_size):
        with self.assertWarns(UserWarning):
            RandAffine(cache_grid=True, spatial_size=spatial_size)

    @parameterized.expand(TEST_CASES_SKIPPED_CONSISTENCY)
    def test_skipped_transform_consistency(self, im, in_dtype):
        t1 = RandAffine(prob=0)
        t2 = RandAffine(prob=1, spatial_size=(10, 11))

        im, *_ = convert_data_type(im, dtype=in_dtype)

        out1 = t1(im)
        out2 = t2(im)

        # check same type
        self.assertEqual(type(out1), type(out2))
        # check matching dtype
        self.assertEqual(out1.dtype, out2.dtype)

    @parameterized.expand(TEST_RANDOMIZE)
    def test_no_randomize(self, initial_randomize, cache_grid):
        rand_affine = RandAffine(
            prob=1,
            rotate_range=(np.pi / 6, 0, 0),
            translate_range=((-2, 2), (-2, 2), (-2, 2)),
            scale_range=((-0.1, 0.1), (-0.1, 0.1), (-0.1, 0.1)),
            spatial_size=(16, 16, 16),
            cache_grid=cache_grid,
            padding_mode="zeros",
        )
        if initial_randomize:
            rand_affine.randomize(None)

        arr = torch.randn((1, 16, 16, 16)) * 100

        arr1 = rand_affine(arr, randomize=False)
        m1 = rand_affine.rand_affine_grid.get_transformation_matrix()

        arr2 = rand_affine(arr, randomize=False)
        m2 = rand_affine.rand_affine_grid.get_transformation_matrix()

        assert_allclose(m1, m2)
        assert_allclose(arr1, arr2)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest
from functools import partial
from typing import Callable

import torch
from parameterized import parameterized

import monai
from monai.bundle import ComponentLocator, ConfigComponent, ConfigExpression, ConfigItem
from monai.data import DataLoader, Dataset
from monai.transforms import LoadImaged, RandTorchVisiond
from monai.utils import min_version, optional_import

_, has_tv = optional_import("torchvision", "0.8.0", min_version)

TEST_CASE_1 = [{"lr": 0.001}, 0.0001]

TEST_CASE_2 = [{"_target_": "LoadImaged", "keys": ["image"], "_desc_": "an image reader for 'image'"}, LoadImaged]
# test full module path
TEST_CASE_3 = [{"_target_": "monai.transforms.LoadImaged", "keys": ["image"]}, LoadImaged]
# test `_disabled_`
TEST_CASE_4 = [{"_target_": "LoadImaged", "_disabled_": True, "keys": ["image"]}, dict]
# test `_disabled_` with string
TEST_CASE_5 = [{"_target_": "LoadImaged", "_disabled_": "true", "keys": ["image"]}, dict]
# test non-monai modules and excludes
TEST_CASE_6 = [{"_target_": "torch.optim.Adam", "params": torch.nn.PReLU().parameters(), "lr": 1e-4}, torch.optim.Adam]
TEST_CASE_7 = [{"_target_": "decollate_batch", "detach": True, "pad": True, "_mode_": "callable"}, partial]
# test args contains "name" field
TEST_CASE_8 = [
    {"_target_": "RandTorchVisiond", "keys": "image", "name": "ColorJitter", "brightness": 0.25},
    RandTorchVisiond,
]
# test execute some function in args, test pre-imported global packages `monai`
TEST_CASE_9 = ["collate_fn", "$monai.data.list_data_collate"]
# test lambda function
TEST_CASE_10 = ["collate_fn", "$lambda x: monai.data.list_data_collate(x) + torch.tensor(var)"]
# test regular expression with reference
TEST_CASE_11 = ["collate_fn", "$var + 100"]


class TestConfigItem(unittest.TestCase):

    @parameterized.expand([TEST_CASE_1])
    def test_item(self, test_input, expected):
        item = ConfigItem(config=test_input)
        conf = item.get_config()
        conf["lr"] = 0.0001
        item.update_config(config=conf)
        self.assertEqual(item.get_config()["lr"], expected)

    @parameterized.expand(
        [TEST_CASE_2, TEST_CASE_3, TEST_CASE_4, TEST_CASE_5, TEST_CASE_6, TEST_CASE_7]
        + ([TEST_CASE_8] if has_tv else [])
    )
    def test_component(self, test_input, output_type):
        locator = ComponentLocator(excludes=["metrics"])
        configer = ConfigComponent(id="test", config=test_input, locator=locator)
        ret = configer.instantiate()
        if test_input.get("_disabled_", False):
            # test `_disabled_` works fine
            self.assertEqual(ret, None)
            return
        self.assertTrue(isinstance(ret, output_type))
        if isinstance(ret, LoadImaged):
            self.assertEqual(ret.keys[0], "image")

    @parameterized.expand([TEST_CASE_9, TEST_CASE_10, TEST_CASE_11])
    def test_expression(self, id, test_input):
        configer = ConfigExpression(id=id, config=test_input, globals={"monai": monai, "torch": torch})
        var = 100
        ret = configer.evaluate(globals={"var": var, "monai": monai})  # `{"monai": monai}` to verify the warning
        if isinstance(ret, Callable):
            self.assertTrue(isinstance(ret([torch.tensor(1), torch.tensor(2)]), torch.Tensor))
        else:
            # also test the `locals` for regular expressions
            ret = configer.evaluate(locals={"var": var})
            self.assertEqual(ret, 200)

    def test_lazy_instantiation(self):
        config = {"_target_": "DataLoader", "dataset": Dataset(data=[1, 2]), "batch_size": 2}
        configer = ConfigComponent(config=config, locator=None)
        init_config = configer.get_config()
        # modify config content at runtime
        init_config["batch_size"] = 4
        configer.update_config(config=init_config)

        ret = configer.instantiate()
        self.assertTrue(isinstance(ret, DataLoader))
        self.assertEqual(ret.batch_size, 4)

    @parameterized.expand([("$import json", "json"), ("$import json as j", "j")])
    def test_import(self, stmt, mod_name):
        test_globals = {}
        ConfigExpression(id="", config=stmt, globals=test_globals).evaluate()
        self.assertTrue(callable(test_globals[mod_name].dump))

    @parameterized.expand(
        [
            ("$from json import dump", "dump"),
            ("$from json import dump, dumps", "dump"),
            ("$from json import dump as jd", "jd"),
            ("$from json import dump as jd, dumps as ds", "jd"),
        ]
    )
    def test_import_from(self, stmt, mod_name):
        test_globals = {}
        ConfigExpression(id="", config=stmt, globals=test_globals).evaluate()
        self.assertTrue(callable(test_globals[mod_name]))
        self.assertTrue(ConfigExpression.is_import_statement(ConfigExpression(id="", config=stmt).config))

    @parameterized.expand(
        [("$from json import dump", True), ("$print()", False), ("$import json", True), ("import json", False)]
    )
    def test_is_import_stmt(self, stmt, expected):
        expr = ConfigExpression(id="", config=stmt)
        flag = expr.is_import_statement(expr.config)
        self.assertEqual(flag, expected)

    def test_error_expr(self):
        with self.assertRaisesRegex(RuntimeError, r"1\+\[\]"):
            ConfigExpression(id="", config="$1+[]").evaluate()


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch

from monai.losses.ssim_loss import SSIMLoss
from monai.utils import set_determinism

# from tests.utils import test_script_save


class TestSSIMLoss(unittest.TestCase):

    def test_shape(self):
        set_determinism(0)
        preds = torch.abs(torch.randn(2, 3, 16, 16))
        target = torch.abs(torch.randn(2, 3, 16, 16))
        preds = preds / preds.max()
        target = target / target.max()

        result = SSIMLoss(spatial_dims=2, data_range=1.0, kernel_type="gaussian", reduction="mean").forward(
            preds, target
        )
        expected_val = 0.9546
        np.testing.assert_allclose(result.detach().cpu().numpy(), expected_val, rtol=1e-4)

        result = SSIMLoss(spatial_dims=2, data_range=1.0, kernel_type="gaussian", reduction="sum").forward(
            preds, target
        )
        expected_val = 1.9092
        np.testing.assert_allclose(result.detach().cpu().numpy(), expected_val, rtol=1e-4)

        result = SSIMLoss(spatial_dims=2, data_range=1.0, kernel_type="gaussian", reduction="none").forward(
            preds, target
        )
        expected_val = [[0.9121], [0.9971]]
        np.testing.assert_allclose(result.detach().cpu().numpy(), expected_val, rtol=1e-4)

    # def test_script(self):
    #     loss = SSIMLoss(spatial_dims=2)
    #     test_input = torch.ones(2, 2, 16, 16)
    #     test_script_save(loss, test_input, test_input)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import sys
import unittest
from copy import deepcopy
from io import StringIO

import numpy as np
import torch
from parameterized import parameterized

import monai.transforms as mt
from monai.data import DataLoader, Dataset
from monai.transforms.compose import execute_compose
from monai.transforms.transform import Randomizable
from monai.utils import set_determinism


def data_from_keys(keys, h, w):
    if keys is None:
        data = torch.arange(h * w).reshape(1, h, w)
    else:
        data = {}
        for i_k, k in enumerate(keys):
            data[k] = torch.arange(h * w).reshape(1, h, w).mul_(i_k * h * w)
    return data


class _RandXform(Randomizable):

    def randomize(self):
        self.val = self.R.random_sample()

    def __call__(self, __unused):
        self.randomize()
        return self.val


class TestCompose(unittest.TestCase):

    def test_empty_compose(self):
        c = mt.Compose()
        i = 1
        self.assertEqual(c(i), 1)

    def test_non_dict_compose(self):

        def a(i):
            return i + "a"

        def b(i):
            return i + "b"

        c = mt.Compose([a, b, a, b])
        self.assertEqual(c(""), "abab")

    def test_dict_compose(self):

        def a(d):
            d = dict(d)
            d["a"] += 1
            return d

        def b(d):
            d = dict(d)
            d["b"] += 1
            return d

        transforms = [a, b, a, b, a]
        data = {"a": 0, "b": 0}
        expected = {"a": 3, "b": 2}

        self.assertDictEqual(mt.Compose(transforms)(data), expected)
        self.assertDictEqual(execute_compose(data, transforms), expected)

    def test_list_dict_compose(self):

        def a(d):  # transform to handle dict data
            d = dict(d)
            d["a"] += 1
            return d

        def b(d):  # transform to generate a batch list of data
            d = dict(d)
            d["b"] += 1
            d = [d] * 5
            return d

        def c(d):  # transform to handle dict data
            d = dict(d)
            d["c"] += 1
            return d

        transforms = [a, a, b, c, c]
        data = {"a": 0, "b": 0, "c": 0}
        expected = {"a": 2, "b": 1, "c": 2}
        value = mt.Compose(transforms)(data)
        for item in value:
            self.assertDictEqual(item, expected)
        value = execute_compose(data, transforms)
        for item in value:
            self.assertDictEqual(item, expected)

    def test_non_dict_compose_with_unpack(self):

        def a(i, i2):
            return i + "a", i2 + "a2"

        def b(i, i2):
            return i + "b", i2 + "b2"

        transforms = [a, b, a, b]
        data = ("", "")
        expected = ("abab", "a2b2a2b2")
        self.assertEqual(mt.Compose(transforms, map_items=False, unpack_items=True)(data), expected)
        self.assertEqual(execute_compose(data, transforms, map_items=False, unpack_items=True), expected)

    def test_list_non_dict_compose_with_unpack(self):

        def a(i, i2):
            return i + "a", i2 + "a2"

        def b(i, i2):
            return i + "b", i2 + "b2"

        transforms = [a, b, a, b]
        data = [("", ""), ("t", "t")]
        expected = [("abab", "a2b2a2b2"), ("tabab", "ta2b2a2b2")]
        self.assertEqual(mt.Compose(transforms, unpack_items=True)(data), expected)
        self.assertEqual(execute_compose(data, transforms, unpack_items=True), expected)

    def test_list_dict_compose_no_map(self):

        def a(d):  # transform to handle dict data
            d = dict(d)
            d["a"] += 1
            return d

        def b(d):  # transform to generate a batch list of data
            d = dict(d)
            d["b"] += 1
            d = [d] * 5
            return d

        def c(d):  # transform to handle dict data
            d = [dict(di) for di in d]
            for di in d:
                di["c"] += 1
            return d

        transforms = [a, a, b, c, c]
        data = {"a": 0, "b": 0, "c": 0}
        expected = {"a": 2, "b": 1, "c": 2}
        value = mt.Compose(transforms, map_items=False)(data)
        for item in value:
            self.assertDictEqual(item, expected)
        value = execute_compose(data, transforms, map_items=False)
        for item in value:
            self.assertDictEqual(item, expected)

    def test_random_compose(self):

        class _Acc(Randomizable):
            self.rand = 0.0

            def randomize(self, data=None):
                self.rand = self.R.rand()

            def __call__(self, data):
                self.randomize()
                return self.rand + data

        c = mt.Compose([_Acc(), _Acc()])
        self.assertNotAlmostEqual(c(0), c(0))
        c.set_random_state(123)
        self.assertAlmostEqual(c(1), 1.61381597)
        c.set_random_state(223)
        c.randomize()
        self.assertAlmostEqual(c(1), 1.90734751)

    def test_randomize_warn(self):

        class _RandomClass(Randomizable):

            def randomize(self, foo1, foo2):
                pass

            def __call__(self, data):
                pass

        c = mt.Compose([_RandomClass(), _RandomClass()])
        with self.assertWarns(Warning):
            c.randomize()

    def test_err_msg(self):
        transforms = mt.Compose([abs, mt.EnsureChannelFirst(), round])
        with self.assertRaisesRegex(Exception, "EnsureChannelFirst"):
            transforms(42.1)

    def test_data_loader(self):
        xform_1 = mt.Compose([_RandXform()])
        train_ds = Dataset([1], transform=xform_1)

        xform_1.set_random_state(123)
        out_1 = train_ds[0]
        self.assertAlmostEqual(out_1, 0.2045649)

        set_determinism(seed=123)
        train_loader = DataLoader(train_ds, num_workers=0)
        out_1 = next(iter(train_loader))
        self.assertAlmostEqual(out_1.cpu().item(), 0.0409280)

        if sys.platform != "win32":  # skip multi-worker tests on win32
            train_loader = DataLoader(train_ds, num_workers=1)
            out_1 = next(iter(train_loader))
            self.assertAlmostEqual(out_1.cpu().item(), 0.78663897075)

            train_loader = DataLoader(train_ds, num_workers=2)
            out_1 = next(iter(train_loader))
            self.assertAlmostEqual(out_1.cpu().item(), 0.785907334)
        set_determinism(None)

    def test_data_loader_2(self):
        set_determinism(seed=123)
        xform_2 = mt.Compose([_RandXform(), _RandXform()])
        train_ds = Dataset([1], transform=xform_2)

        out_2 = train_ds[0]
        self.assertAlmostEqual(out_2, 0.4092510)

        train_loader = DataLoader(train_ds, num_workers=0)
        out_2 = next(iter(train_loader))
        self.assertAlmostEqual(out_2.cpu().item(), 0.98921915918)

        if sys.platform != "win32":  # skip multi-worker tests on win32
            train_loader = DataLoader(train_ds, num_workers=1)
            out_2 = next(iter(train_loader))
            self.assertAlmostEqual(out_2.cpu().item(), 0.32985207)

            train_loader = DataLoader(train_ds, num_workers=2)
            out_1 = next(iter(train_loader))
            self.assertAlmostEqual(out_1.cpu().item(), 0.28602141572)
        set_determinism(None)

    def test_flatten_and_len(self):
        x = mt.EnsureChannelFirst(channel_dim="no_channel")
        t1 = mt.Compose([x, x, x, x, mt.Compose([mt.Compose([x, x]), x, x])])

        t2 = t1.flatten()
        for t in t2.transforms:
            self.assertNotIsInstance(t, mt.Compose)

        # test len
        self.assertEqual(len(t1), 8)

    def test_backwards_compatible_imports(self):
        from monai.transforms.transform import MapTransform, RandomizableTransform, Transform  # noqa: F401


TEST_COMPOSE_EXECUTE_TEST_CASES = [
    [None, tuple()],
    [None, (mt.Rotate(np.pi / 8),)],
    [None, (mt.Flip(0), mt.Flip(1), mt.Rotate90(1), mt.Zoom(0.8), mt.NormalizeIntensity())],
    [("a",), (mt.Rotated(("a",), np.pi / 8),)],
]


class TestComposeExecute(unittest.TestCase):

    @parameterized.expand(TEST_COMPOSE_EXECUTE_TEST_CASES)
    def test_compose_execute_equivalence(self, keys, pipeline):
        data = data_from_keys(keys, 12, 16)

        expected = mt.Compose(deepcopy(pipeline))(data)

        for cutoff in range(len(pipeline)):
            c = mt.Compose(deepcopy(pipeline))
            actual = c(c(data, end=cutoff), start=cutoff)
            if isinstance(actual, dict):
                for k in actual.keys():
                    self.assertTrue(torch.allclose(expected[k], actual[k]))
            else:
                self.assertTrue(torch.allclose(expected, actual))

            p = deepcopy(pipeline)
            actual = execute_compose(execute_compose(data, p, start=0, end=cutoff), p, start=cutoff)
            if isinstance(actual, dict):
                for k in actual.keys():
                    self.assertTrue(torch.allclose(expected[k], actual[k]))
            else:
                self.assertTrue(torch.allclose(expected, actual))

    @parameterized.expand(TEST_COMPOSE_EXECUTE_TEST_CASES)
    def test_compose_execute_bad_start_param(self, keys, pipeline):
        data = data_from_keys(keys, 12, 16)

        c = mt.Compose(deepcopy(pipeline))
        with self.assertRaises(ValueError):
            c(data, start=None)
        with self.assertRaises(ValueError):
            c(data, start=None)

        with self.assertRaises(ValueError):
            execute_compose(data, deepcopy(pipeline), start=None)

        c = mt.Compose(deepcopy(pipeline))
        with self.assertRaises(ValueError):
            c(data, start=-1)
        with self.assertRaises(ValueError):
            c(data, start=-1)

        with self.assertRaises(ValueError):
            execute_compose(data, deepcopy(pipeline), start=-1)

    @parameterized.expand(TEST_COMPOSE_EXECUTE_TEST_CASES)
    def test_compose_execute_negative_range(self, keys, pipeline):
        data = data_from_keys(keys, 12, 16)

        with self.assertRaises(ValueError):
            c = mt.Compose(deepcopy(pipeline))
            c(data, start=2, end=1)

        with self.assertRaises(ValueError):
            execute_compose(data, deepcopy(pipeline), start=2, end=1)

    @parameterized.expand(TEST_COMPOSE_EXECUTE_TEST_CASES)
    def test_compose_execute_bad_end_param(self, keys, pipeline):
        data = data_from_keys(keys, 12, 16)

        with self.assertRaises(ValueError):
            c = mt.Compose(deepcopy(pipeline))
            c(data, end=len(pipeline) + 1)

        with self.assertRaises(ValueError):
            execute_compose(data, deepcopy(pipeline), end=len(pipeline) + 1)

    @parameterized.expand(TEST_COMPOSE_EXECUTE_TEST_CASES)
    def test_compose_execute_empty_range(self, keys, pipeline):
        data = data_from_keys(keys, 12, 16)

        c = mt.Compose(deepcopy(pipeline))
        for i in range(len(pipeline)):
            result = c(data, start=i, end=i)
            self.assertIs(data, result)

    @parameterized.expand(TEST_COMPOSE_EXECUTE_TEST_CASES)
    def test_compose_with_logger(self, keys, pipeline):
        data = data_from_keys(keys, 12, 16)

        c = mt.Compose(deepcopy(pipeline), log_stats="a_logger_name")
        c(data)


TEST_COMPOSE_EXECUTE_LOGGING_TEST_CASES = [
    [
        None,
        (mt.Flip(0), mt.Spacing((1.2, 1.2)), mt.Flip(1), mt.Rotate90(1), mt.Zoom(0.8), mt.NormalizeIntensity()),
        False,
        (
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'Flip', transform.lazy: False\n"
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'Spacing', transform.lazy: False\n"
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'Flip', transform.lazy: False\n"
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'Rotate90', transform.lazy: False\n"
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'Zoom', transform.lazy: False\n"
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'NormalizeIntensity', transform is not lazy\n"
        ),
    ],
    [
        None,
        (
            mt.Flip(0, lazy=True),
            mt.Spacing((1.2, 1.2), lazy=True),
            mt.Flip(1, lazy=True),
            mt.Rotate90(1),
            mt.Zoom(0.8, lazy=True),
            mt.NormalizeIntensity(),
        ),
        None,
        (
            "INFO - Accumulate pending transforms - lazy: None, pending: 0, "
            "upcoming 'Flip', transform.lazy: True\n"
            "INFO - Accumulate pending transforms - lazy: None, pending: 1, "
            "upcoming 'Spacing', transform.lazy: True\n"
            "INFO - Accumulate pending transforms - lazy: None, pending: 2, "
            "upcoming 'Flip', transform.lazy: True\n"
            "INFO - Apply pending transforms - lazy: None, pending: 3, "
            "upcoming 'Rotate90', transform.lazy: False\n"
            "INFO - Pending transforms applied: applied_operations: 3\n"
            "INFO - Accumulate pending transforms - lazy: None, pending: 0, "
            "upcoming 'Zoom', transform.lazy: True\n"
            "INFO - Apply pending transforms - lazy: None, pending: 1, "
            "upcoming 'NormalizeIntensity', transform is not lazy\n"
            "INFO - Pending transforms applied: applied_operations: 5\n"
        ),
    ],
    [
        None,
        (mt.Flip(0), mt.Spacing((1.2, 1.2)), mt.Flip(1), mt.Rotate90(1), mt.Zoom(0.8), mt.NormalizeIntensity()),
        True,
        (
            "INFO - Accumulate pending transforms - lazy: True, pending: 0, "
            "upcoming 'Flip', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, pending: 1, "
            "upcoming 'Spacing', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, pending: 2, "
            "upcoming 'Flip', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, pending: 3, "
            "upcoming 'Rotate90', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, pending: 4, "
            "upcoming 'Zoom', transform.lazy: False (overridden)\n"
            "INFO - Apply pending transforms - lazy: True, pending: 5, "
            "upcoming 'NormalizeIntensity', transform is not lazy\n"
            "INFO - Pending transforms applied: applied_operations: 5\n"
        ),
    ],
    [
        ("a", "b"),
        (
            mt.Flipd(("a", "b"), 0),
            mt.Spacingd(("a", "b"), 1.2),
            mt.Rotate90d(("a", "b"), 1),
            mt.NormalizeIntensityd(("a",)),
        ),
        True,
        (
            "INFO - Accumulate pending transforms - lazy: True, key: 'a', pending: 0, "
            "upcoming 'Flipd', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'b', pending: 0, "
            "upcoming 'Flipd', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'a', pending: 1, "
            "upcoming 'Spacingd', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'b', pending: 1, "
            "upcoming 'Spacingd', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'a', pending: 2, "
            "upcoming 'Rotate90d', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'b', pending: 2, "
            "upcoming 'Rotate90d', transform.lazy: False (overridden)\n"
            "INFO - Apply pending transforms - lazy: True, key: 'a', pending: 3, "
            "upcoming 'NormalizeIntensityd', transform is not lazy\n"
            "INFO - Pending transforms applied: key: 'a', applied_operations: 3\n"
            "INFO - Pending transforms applied: key: 'b', applied_operations: 3\n"
        ),
    ],
    [
        ("a", "b"),
        (
            mt.Flipd(keys="a", spatial_axis=0),
            mt.Rotate90d(keys="b", k=1, allow_missing_keys=True),
            mt.Zoomd(keys=("a", "b"), zoom=0.8, allow_missing_keys=True),
            mt.Spacingd(keys="a", pixdim=1.2),
        ),
        True,
        (
            "INFO - Accumulate pending transforms - lazy: True, key: 'a', pending: 0, "
            "upcoming 'Flipd', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'b', pending: 0, "
            "upcoming 'Rotate90d', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'a', pending: 1, "
            "upcoming 'Zoomd', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'b', pending: 1, "
            "upcoming 'Zoomd', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'a', pending: 2, "
            "upcoming 'Spacingd', transform.lazy: False (overridden)\n"
            "INFO - Pending transforms applied: key: 'a', applied_operations: 3\n"
            "INFO - Pending transforms applied: key: 'b', applied_operations: 2\n"
        ),
    ],
    [
        None,
        (
            mt.Flip(0),
            mt.Spacing((1.2, 1.2)),
            mt.Flip(1),
            mt.ApplyPending(),
            mt.Rotate90(1),
            mt.Zoom(0.8),
            mt.NormalizeIntensity(),
        ),
        False,
        (
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'Flip', transform.lazy: False\n"
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'Spacing', transform.lazy: False\n"
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'Flip', transform.lazy: False\n"
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'ApplyPending', transform is not lazy\n"
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'Rotate90', transform.lazy: False\n"
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'Zoom', transform.lazy: False\n"
            "INFO - Apply pending transforms - lazy: False, pending: 0, "
            "upcoming 'NormalizeIntensity', transform is not lazy\n"
        ),
    ],
    [
        None,
        (
            mt.Flip(0),
            mt.Spacing((1.2, 1.2)),
            mt.Flip(1),
            mt.ApplyPending(),
            mt.Rotate90(1),
            mt.Zoom(0.8),
            mt.NormalizeIntensity(),
        ),
        True,
        (
            "INFO - Accumulate pending transforms - lazy: True, pending: 0, "
            "upcoming 'Flip', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, pending: 1, "
            "upcoming 'Spacing', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, pending: 2, "
            "upcoming 'Flip', transform.lazy: False (overridden)\n"
            "INFO - Apply pending transforms - lazy: True, pending: 3, "
            "upcoming 'ApplyPending', transform is not lazy\n"
            "INFO - Pending transforms applied: applied_operations: 3\n"
            "INFO - Accumulate pending transforms - lazy: True, pending: 0, "
            "upcoming 'Rotate90', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, pending: 1, "
            "upcoming 'Zoom', transform.lazy: False (overridden)\n"
            "INFO - Apply pending transforms - lazy: True, pending: 2, "
            "upcoming 'NormalizeIntensity', transform is not lazy\n"
            "INFO - Pending transforms applied: applied_operations: 5\n"
        ),
    ],
    [
        ("a", "b"),
        (
            mt.Flipd(keys="a", spatial_axis=0),
            mt.Rotate90d(keys="b", k=1, allow_missing_keys=True),
            mt.ApplyPendingd(keys=("a", "b")),
            mt.Zoomd(keys=("a", "b"), zoom=0.8, allow_missing_keys=True),
            mt.Spacingd(keys="a", pixdim=1.2),
        ),
        True,
        (
            "INFO - Accumulate pending transforms - lazy: True, key: 'a', pending: 0, "
            "upcoming 'Flipd', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'b', pending: 0, "
            "upcoming 'Rotate90d', transform.lazy: False (overridden)\n"
            "INFO - Apply pending transforms - lazy: True, key: 'a', pending: 1, "
            "upcoming 'ApplyPendingd', transform is not lazy\n"
            "INFO - Apply pending transforms - lazy: True, key: 'b', pending: 1, "
            "upcoming 'ApplyPendingd', transform is not lazy\n"
            "INFO - Pending transforms applied: key: 'a', applied_operations: 1\n"
            "INFO - Pending transforms applied: key: 'b', applied_operations: 1\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'a', pending: 0, "
            "upcoming 'Zoomd', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'b', pending: 0, "
            "upcoming 'Zoomd', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, key: 'a', pending: 1, "
            "upcoming 'Spacingd', transform.lazy: False (overridden)\n"
            "INFO - Pending transforms applied: key: 'a', applied_operations: 3\n"
            "INFO - Pending transforms applied: key: 'b', applied_operations: 2\n"
        ),
    ],
]

TEST_COMPOSE_LAZY_ON_CALL_LOGGING_TEST_CASES = [
    [
        mt.Compose,
        (mt.Flip(0), mt.Spacing((1.2, 1.2))),
        True,
        (
            "INFO - Accumulate pending transforms - lazy: True, pending: 0, "
            "upcoming 'Flip', transform.lazy: False (overridden)\n"
            "INFO - Accumulate pending transforms - lazy: True, pending: 1, "
            "upcoming 'Spacing', transform.lazy: False (overridden)\n"
            "INFO - Pending transforms applied: applied_operations: 2\n"
        ),
    ],
    [
        mt.SomeOf,
        (mt.Flip(0),),
        True,
        (
            "INFO - Accumulate pending transforms - lazy: True, pending: 0, "
            "upcoming 'Flip', transform.lazy: False (overridden)\n"
            "INFO - Pending transforms applied: applied_operations: 1\n"
        ),
    ],
    [
        mt.RandomOrder,
        (mt.Flip(0),),
        True,
        (
            "INFO - Accumulate pending transforms - lazy: True, pending: 0, "
            "upcoming 'Flip', transform.lazy: False (overridden)\n"
            "INFO - Pending transforms applied: applied_operations: 1\n"
        ),
    ],
    [
        mt.OneOf,
        (mt.Flip(0),),
        True,
        (
            "INFO - Accumulate pending transforms - lazy: True, pending: 0, "
            "upcoming 'Flip', transform.lazy: False (overridden)\n"
            "INFO - Pending transforms applied: applied_operations: 1\n"
        ),
    ],
    [
        mt.OneOf,
        (mt.Flip(0),),
        False,
        ("INFO - Apply pending transforms - lazy: False, pending: 0, " "upcoming 'Flip', transform.lazy: False\n"),
    ],
]


class TestComposeExecuteWithLogging(unittest.TestCase):
    LOGGER_NAME = "a_logger_name"

    def init_logger(self, name=LOGGER_NAME):
        stream = StringIO()
        handler = logging.StreamHandler(stream)
        formatter = logging.Formatter("%(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        logger = logging.getLogger(name)
        logger.setLevel(logging.INFO)
        while len(logger.handlers) > 0:
            logger.removeHandler(logger.handlers[-1])
        logger.addHandler(handler)
        return handler, stream

    @parameterized.expand(TEST_COMPOSE_EXECUTE_LOGGING_TEST_CASES)
    def test_compose_with_logging(self, keys, pipeline, lazy, expected):
        handler, stream = self.init_logger(name=self.LOGGER_NAME)

        data = data_from_keys(keys, 12, 16)
        c = mt.Compose(deepcopy(pipeline), lazy=lazy, log_stats=self.LOGGER_NAME)
        c(data)

        handler.flush()
        actual = stream.getvalue()
        self.assertEqual(actual, expected)

    @parameterized.expand(TEST_COMPOSE_LAZY_ON_CALL_LOGGING_TEST_CASES)
    def test_compose_lazy_on_call_with_logging(self, compose_type, pipeline, lazy_on_call, expected):
        handler, stream = self.init_logger(name=self.LOGGER_NAME)

        data = data_from_keys(None, 12, 16)
        c = compose_type(deepcopy(pipeline), log_stats=self.LOGGER_NAME)
        c(data, lazy=lazy_on_call)

        handler.flush()
        actual = stream.getvalue()
        self.assertEqual(actual, expected)


class TestOps:

    @staticmethod
    def concat(value):

        def _inner(data):
            return data + value

        return _inner

    @staticmethod
    def concatd(value):

        def _inner(data):
            return {k: v + value for k, v in data.items()}

        return _inner

    @staticmethod
    def concata(value):

        def _inner(data1, data2):
            return data1 + value, data2 + value

        return _inner


TEST_COMPOSE_EXECUTE_FLAG_TEST_CASES = [
    [{}, ("",), (TestOps.concat("a"), TestOps.concat("b"))],
    [{"unpack_items": True}, ("x", "y"), (TestOps.concat("a"), TestOps.concat("b"))],
    [{"map_items": False}, {"x": "1", "y": "2"}, (TestOps.concatd("a"), TestOps.concatd("b"))],
    [{"unpack_items": True, "map_items": False}, ("x", "y"), (TestOps.concata("a"), TestOps.concata("b"))],
]


class TestComposeExecuteWithFlags(unittest.TestCase):

    @parameterized.expand(TEST_COMPOSE_EXECUTE_FLAG_TEST_CASES)
    def test_compose_execute_equivalence_with_flags(self, flags, data, pipeline):
        expected = mt.Compose(pipeline, **flags)(data)

        for cutoff in range(len(pipeline)):
            c = mt.Compose(deepcopy(pipeline), **flags)
            actual = c(c(data, end=cutoff), start=cutoff)
            if isinstance(actual, dict):
                for k in actual.keys():
                    self.assertEqual(expected[k], actual[k])
            else:
                self.assertEqual(expected, actual)

            p = deepcopy(pipeline)
            actual = execute_compose(execute_compose(data, p, start=0, end=cutoff, **flags), p, start=cutoff, **flags)
            if isinstance(actual, dict):
                for k in actual.keys():
                    self.assertEqual(expected[k], actual[k])
            else:
                self.assertEqual(expected, actual)


class TestComposeCallableInput(unittest.TestCase):

    def test_value_error_when_not_sequence(self):
        data = torch.tensor(np.random.randn(1, 5, 5))

        xform = mt.Compose([mt.Flip(0), mt.Flip(0)])
        res = xform(data)
        np.testing.assert_allclose(data, res, atol=1e-3)

        with self.assertRaises(ValueError):
            mt.Compose(mt.Flip(0), mt.Flip(0))(data)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.data.utils import dev_collate

TEST_CASES = [
    [
        [
            {"img": 2, "meta": {"shape": [torch.tensor(1.0)]}},
            {"img": 3, "meta": {"shape": [np.asarray(1.0)]}},
            {"img": 4, "meta": {"shape": [torch.tensor(1.0)]}},
        ],
        "got numpy.ndarray",
    ],
    [[["img", np.array([2])], ["img", np.array([3, 4])], ["img", np.array([4])]], "size"],
    [[["img", [2]], ["img", [3, 4]], ["img", 4]], "type"],
    [[["img", [2, 2]], ["img", [3, 4]], ["img", 4]], "type"],
]


class DevCollateTest(unittest.TestCase):

    @parameterized.expand(TEST_CASES)
    def test_dev_collate(self, inputs, msg):
        with self.assertLogs(level=logging.CRITICAL) as log:
            dev_collate(inputs)
            self.assertRegex(" ".join(log.output), f"{msg}")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.transforms import CastToType
from monai.utils import optional_import
from monai.utils.type_conversion import get_equivalent_dtype
from tests.utils import HAS_CUPY, TEST_NDARRAYS

cp, _ = optional_import("cupy")

TESTS = []
for p in TEST_NDARRAYS:
    for out_dtype in (np.float64, torch.float64):
        TESTS.append([out_dtype, p(np.array([[0, 1], [1, 2]], dtype=np.float32)), out_dtype])

TESTS_CUPY = [
    [np.float32, np.array([[0, 1], [1, 2]], dtype=np.float32), np.float32],
    [np.float32, np.array([[0, 1], [1, 2]], dtype=np.uint8), np.float32],
    [np.uint8, np.array([[0, 1], [1, 2]], dtype=np.float32), np.uint8],
]


class TestCastToType(unittest.TestCase):

    @parameterized.expand(TESTS)
    def test_type(self, out_dtype, input_data, expected_type):
        result = CastToType(dtype=out_dtype)(input_data)
        self.assertEqual(result.dtype, get_equivalent_dtype(expected_type, type(result)))

        result = CastToType()(input_data, out_dtype)
        self.assertEqual(result.dtype, get_equivalent_dtype(expected_type, type(result)))

    @parameterized.expand(TESTS_CUPY)
    @unittest.skipUnless(HAS_CUPY, "Requires CuPy")
    def test_type_cupy(self, out_dtype, input_data, expected_type):
        input_data = cp.asarray(input_data)

        result = CastToType(dtype=out_dtype)(input_data)
        self.assertTrue(isinstance(result, cp.ndarray))
        self.assertEqual(result.dtype, get_equivalent_dtype(expected_type, type(result)))

        result = CastToType()(input_data, out_dtype)
        self.assertTrue(isinstance(result, cp.ndarray))
        self.assertEqual(result.dtype, get_equivalent_dtype(expected_type, type(result)))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks import eval_mode, train_mode
from monai.networks.nets import HoVerNet
from monai.networks.nets.hovernet import _DenseLayerDecoder
from tests.utils import test_script_save

device = "cuda" if torch.cuda.is_available() else "cpu"

TEST_CASE_0 = [  # fast mode
    {"out_classes": 5, "mode": HoVerNet.Mode.FAST},
    (1, 3, 256, 256),
    {HoVerNet.Branch.NP: (1, 2, 164, 164), HoVerNet.Branch.NC: (1, 5, 164, 164), HoVerNet.Branch.HV: (1, 2, 164, 164)},
]

TEST_CASE_1 = [  # original mode
    {"out_classes": 6, "mode": HoVerNet.Mode.ORIGINAL},
    (1, 3, 270, 270),
    {HoVerNet.Branch.NP: (1, 2, 80, 80), HoVerNet.Branch.NC: (1, 6, 80, 80), HoVerNet.Branch.HV: (1, 2, 80, 80)},
]

TEST_CASE_2 = [  # dropout
    {"mode": HoVerNet.Mode.FAST, "dropout_prob": 0.5, "out_classes": 3},
    (1, 3, 256, 256),
    {HoVerNet.Branch.NP: (1, 2, 164, 164), HoVerNet.Branch.NC: (1, 3, 164, 164), HoVerNet.Branch.HV: (1, 2, 164, 164)},
]

TEST_CASE_3 = [  # np_out_channels
    {"mode": HoVerNet.Mode.FAST, "np_out_channels": 3, "out_classes": 2},
    (1, 3, 256, 256),
    {HoVerNet.Branch.NP: (1, 3, 164, 164), HoVerNet.Branch.NC: (1, 2, 164, 164), HoVerNet.Branch.HV: (1, 2, 164, 164)},
]

CASES = [TEST_CASE_0, TEST_CASE_1, TEST_CASE_2, TEST_CASE_3]

ILL_CASES = [
    [{"out_classes": 6, "mode": 3}],
    [{"out_classes": 6, "mode": "Wrong"}],
    [{"out_classes": 1000, "mode": HoVerNet.Mode.ORIGINAL}],
    [{"out_classes": 1, "mode": HoVerNet.Mode.ORIGINAL}],
    [{"out_classes": 6, "mode": HoVerNet.Mode.ORIGINAL, "dropout_prob": 100}],
]


def check_branch(branch, mode):
    if mode == HoVerNet.Mode.ORIGINAL:
        ksize = 5
    else:
        ksize = 3

    if branch.decoderblock1.conva.kernel_size != (ksize, ksize):
        return True
    if branch.decoderblock1.convf.kernel_size != (1, 1):
        return True
    for block in branch.decoderblock1:
        if isinstance(block, _DenseLayerDecoder):
            if block.layers.conv1.kernel_size != (1, 1) or block.layers.conv2.kernel_size != (ksize, ksize):
                return True

    if branch.decoderblock2.conva.kernel_size != (ksize, ksize):
        return True
    if branch.decoderblock2.convf.kernel_size != (1, 1):
        return True

    for block in branch.decoderblock2:
        if isinstance(block, _DenseLayerDecoder):
            if block.layers.conv1.kernel_size != (1, 1) or block.layers.conv2.kernel_size != (ksize, ksize):
                return True

    return False


def check_output(out_block, mode):
    if mode == HoVerNet.Mode.ORIGINAL:
        ksize = 5
    else:
        ksize = 3

    if out_block.decoderblock3.conva.kernel_size != (ksize, ksize) or out_block.decoderblock3.conva.stride != (1, 1):
        return True
    if out_block.decoderblock4.conv.kernel_size != (1, 1) or out_block.decoderblock4.conv.stride != (1, 1):
        return True

    return False


def check_kernels(net, mode):
    # Check the Encoder blocks
    for layer_num, res_block in enumerate(net.res_blocks):
        for inner_num, layer in enumerate(res_block.layers):
            if layer_num > 0 and inner_num == 0:
                sz = 2
            else:
                sz = 1

            if (
                layer.layers.conv1.kernel_size != (1, 1)
                or layer.layers.conv2.kernel_size != (3, 3)
                or layer.layers.conv3.kernel_size != (1, 1)
            ):
                return True

            if (
                layer.layers.conv1.stride != (1, 1)
                or layer.layers.conv2.stride != (sz, sz)
                or layer.layers.conv3.stride != (1, 1)
            ):
                return True

        sz2 = 1
        if layer_num > 0:
            sz2 = 2
        if res_block.shortcut.kernel_size != (1, 1) or res_block.shortcut.stride != (sz2, sz2):
            return True

    if net.bottleneck.conv_bottleneck.kernel_size != (1, 1) or net.bottleneck.conv_bottleneck.stride != (1, 1):
        return True

    # Check HV Branch
    if check_branch(net.horizontal_vertical.decoder_blocks, mode):
        return True
    if check_output(net.horizontal_vertical.output_features, mode):
        return True

    # Check NP Branch
    if check_branch(net.nucleus_prediction.decoder_blocks, mode):
        return True
    if check_output(net.nucleus_prediction.output_features, mode):
        return True

    # Check NC Branch
    if check_branch(net.type_prediction.decoder_blocks, mode):
        return True
    if check_output(net.type_prediction.output_features, mode):
        return True

    return False


class TestHoverNet(unittest.TestCase):

    @parameterized.expand(CASES)
    def test_shape(self, input_param, input_shape, expected_shapes):
        input_param["decoder_padding"] = False
        net = HoVerNet(**input_param).to(device)
        with eval_mode(net):
            result = net.forward(torch.randn(input_shape).to(device))
            for item in result:
                self.assertEqual(result[item].shape, expected_shapes[item])

    @parameterized.expand(CASES)
    def test_decoder_padding_shape(self, input_param, input_shape, expected_shapes):
        if input_param["mode"] == HoVerNet.Mode.FAST:
            input_param["decoder_padding"] = True
            net = HoVerNet(**input_param).to(device)
            with eval_mode(net):
                result = net.forward(torch.randn(input_shape).to(device))
                for item in result:
                    expected_shape = expected_shapes[item]
                    padding_expected_shape = list(expected_shape)
                    padding_expected_shape[2:] = input_shape[2:]
                    self.assertEqual(result[item].shape, tuple(padding_expected_shape))
        else:
            pass

    def test_script(self):
        for padding_flag in [True, False]:
            net = HoVerNet(mode=HoVerNet.Mode.FAST, decoder_padding=padding_flag)
        test_data = torch.randn(1, 3, 256, 256)
        test_script_save(net, test_data)

    def test_ill_input_shape(self):
        net = HoVerNet(mode=HoVerNet.Mode.FAST)
        with eval_mode(net):
            with self.assertRaises(ValueError):
                net.forward(torch.randn(1, 3, 270, 260))

    def test_kernels_strides(self):
        net = HoVerNet(mode=HoVerNet.Mode.FAST, out_classes=2)
        with eval_mode(net):
            self.assertEqual(check_kernels(net, HoVerNet.Mode.FAST), False)

        net = HoVerNet(mode=HoVerNet.Mode.ORIGINAL, out_classes=2)
        with eval_mode(net):
            self.assertEqual(check_kernels(net, HoVerNet.Mode.ORIGINAL), False)

    def test_freeze_encoder(self):
        net = HoVerNet(mode=HoVerNet.Mode.FAST, freeze_encoder=True)
        with train_mode(net):
            for _, param in net.res_blocks[1:].named_parameters():
                self.assertFalse(param.requires_grad)
            for name, param in net.res_blocks[0].named_parameters():
                if param.requires_grad is True:
                    self.assertTrue("bna_block" or "shortcut" in name)

    @parameterized.expand(ILL_CASES)
    def test_ill_input_hyper_params(self, input_param):
        with self.assertRaises(ValueError):
            _ = HoVerNet(**input_param)


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.data import MetaTensor, list_data_collate

a = {"image": np.array([1, 2, 3]), "label": MetaTensor([4, 5, 6])}
b = {"image": np.array([7, 8, 9]), "label": MetaTensor([10, 11, 12])}
c = {"image": np.array([13, 14, 15]), "label": MetaTensor([16, 7, 18])}
d = {"image": np.array([19, 20, 21]), "label": MetaTensor([22, 23, 24])}
TEST_CASE_1 = [[[a, b], [c, d]], dict, torch.Size([4, 3])]  # dataset returns a list of dictionary data

e = (np.array([1, 2, 3]), MetaTensor([4, 5, 6]))
f = (np.array([7, 8, 9]), MetaTensor([10, 11, 12]))
g = (np.array([13, 14, 15]), MetaTensor([16, 7, 18]))
h = (np.array([19, 20, 21]), MetaTensor([22, 23, 24]))
TEST_CASE_2 = [[[e, f], [g, h]], list, torch.Size([4, 3])]  # dataset returns a list of tuple data

g_m = (np.array([13, 14, 15]), MetaTensor([16, 7, 18], meta={"key1": 0}))
h_m = (np.array([19, 20, 21]), MetaTensor([22, 23, 24], meta={"key2": 1}))
TEST_CASE_3 = [[[g_m], [h_m]], list, torch.Size([2, 3])]


class TestListDataCollate(unittest.TestCase):

    @parameterized.expand([TEST_CASE_1, TEST_CASE_2, TEST_CASE_3])
    def test_type_shape(self, input_data, expected_type, expected_shape):
        result = list_data_collate(input_data)
        self.assertIsInstance(result, expected_type)
        if isinstance(result, dict):
            image = result["image"]
            label = result["label"]
        else:
            image = result[0]
            label = result[1]
        self.assertEqual(image.shape, expected_shape)
        self.assertEqual(label.shape, expected_shape)
        self.assertTrue(isinstance(label, MetaTensor))
        self.assertTrue(label.is_batch, True)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import random
import sys
import unittest
from copy import deepcopy
from functools import partial
from typing import TYPE_CHECKING
from unittest.case import skipUnless

import numpy as np
import torch
from parameterized import parameterized

from monai.data import CacheDataset, DataLoader, MetaTensor, create_test_image_2d, create_test_image_3d, decollate_batch
from monai.networks.nets import UNet
from monai.transforms import (
    Affined,
    BorderPadd,
    CenterScaleCropd,
    CenterSpatialCropd,
    Compose,
    CropForegroundd,
    DivisiblePadd,
    EnsureChannelFirstd,
    Flipd,
    FromMetaTensord,
    InvertibleTransform,
    Lambdad,
    LoadImaged,
    Orientationd,
    RandAffined,
    RandAxisFlipd,
    RandCropByLabelClassesd,
    RandCropByPosNegLabeld,
    RandFlipd,
    RandLambdad,
    Randomizable,
    RandRotate90d,
    RandRotated,
    RandSpatialCropd,
    RandSpatialCropSamplesd,
    RandWeightedCropd,
    RandZoomd,
    Resized,
    ResizeWithPadOrCrop,
    ResizeWithPadOrCropd,
    Rotate90d,
    Rotated,
    Spacingd,
    SpatialCropd,
    SpatialPadd,
    ToMetaTensord,
    Transposed,
    Zoomd,
    allow_missing_keys_mode,
    convert_applied_interp_mode,
    reset_ops_id,
)
from monai.utils import first, get_seed, optional_import, set_determinism
from tests.utils import make_nifti_image, make_rand_affine

if TYPE_CHECKING:
    has_nib = True
else:
    _, has_nib = optional_import("nibabel")

KEYS = ["image", "label"]

TESTS: list[tuple] = []

# For pad, start with odd/even images and add odd/even amounts
for name in ("1D even", "1D odd"):
    for val in (3, 4):
        for t in (
            partial(SpatialPadd, spatial_size=val, method="symmetric"),
            partial(SpatialPadd, spatial_size=val, method="end"),
            partial(BorderPadd, spatial_border=[val, val + 1]),
            partial(DivisiblePadd, k=val),
            partial(ResizeWithPadOrCropd, spatial_size=20 + val),
            partial(CenterSpatialCropd, roi_size=10 + val),
            partial(CenterScaleCropd, roi_scale=0.8),
            partial(CropForegroundd, source_key="label"),
            partial(SpatialCropd, roi_center=10, roi_size=10 + val),
            partial(SpatialCropd, roi_center=11, roi_size=10 + val),
            partial(SpatialCropd, roi_start=val, roi_end=17),
            partial(SpatialCropd, roi_start=val, roi_end=16),
            partial(RandSpatialCropd, roi_size=12 + val),
            partial(ResizeWithPadOrCropd, spatial_size=21 - val),
        ):
            TESTS.append((t.func.__name__ + name, name, 0, True, t(KEYS)))  # type: ignore

# non-sensical tests: crop bigger or pad smaller or -ve values
for t in (
    partial(DivisiblePadd, k=-3),
    partial(CenterSpatialCropd, roi_size=-3),
    partial(RandSpatialCropd, roi_size=-3),
    partial(SpatialPadd, spatial_size=15),
    partial(BorderPadd, spatial_border=[15, 16]),
    partial(CenterSpatialCropd, roi_size=30),
    partial(SpatialCropd, roi_center=10, roi_size=100),
    partial(SpatialCropd, roi_start=3, roi_end=100),
):
    TESTS.append((t.func.__name__ + "bad 1D even", "1D even", 0, True, t(KEYS)))  # type: ignore

TESTS.append(
    (
        "SpatialPadd (x2) 2d",
        "2D",
        0,
        True,
        SpatialPadd(KEYS, spatial_size=[111, 113], method="end"),
        SpatialPadd(KEYS, spatial_size=[118, 117]),
    )
)

TESTS.append(("SpatialPadd 3d", "3D", 0, True, SpatialPadd(KEYS, spatial_size=[112, 113, 116])))

TESTS.append(("SpatialCropd 2d", "2D", 0, True, SpatialCropd(KEYS, [49, 51], [90, 89])))

TESTS.append(
    (
        "SpatialCropd 3d",
        "3D",
        0,
        True,
        SpatialCropd(KEYS, roi_slices=[slice(s, e) for s, e in zip([None, None, -99], [None, -2, None])]),
    )
)

TESTS.append(("SpatialCropd 2d", "2D", 0, True, SpatialCropd(KEYS, [49, 51], [390, 89])))

TESTS.append(("SpatialCropd 3d", "3D", 0, True, SpatialCropd(KEYS, [49, 51, 44], [90, 89, 93])))

TESTS.append(("RandSpatialCropd 2d", "2D", 0, True, RandSpatialCropd(KEYS, [96, 93], None, True, False)))

TESTS.append(("RandSpatialCropd 3d", "3D", 0, True, RandSpatialCropd(KEYS, [96, 93, 92], None, False, False)))

TESTS.append(("BorderPadd 2d", "2D", 0, True, BorderPadd(KEYS, [3, 7, 2, 5])))

TESTS.append(("BorderPadd 2d", "2D", 0, True, BorderPadd(KEYS, [3, 7])))

TESTS.append(("BorderPadd 3d", "3D", 0, True, BorderPadd(KEYS, [4])))

TESTS.append(("DivisiblePadd 2d", "2D", 0, True, DivisiblePadd(KEYS, k=4)))

TESTS.append(("DivisiblePadd 3d", "3D", 0, True, DivisiblePadd(KEYS, k=[4, 8, 11])))

TESTS.append(("CenterSpatialCropd 2d", "2D", 0, True, CenterSpatialCropd(KEYS, roi_size=95)))

TESTS.append(("CenterSpatialCropd 3d", "3D", 0, True, CenterSpatialCropd(KEYS, roi_size=[95, 97, 98])))

TESTS.append(("CropForegroundd 2d", "2D", 0, True, CropForegroundd(KEYS, source_key="label", margin=2)))

TESTS.append(("CropForegroundd 3d", "3D", 0, True, CropForegroundd(KEYS, source_key="label", k_divisible=[5, 101, 2])))

TESTS.append(("ResizeWithPadOrCropd 3d", "3D", 0, True, ResizeWithPadOrCropd(KEYS, [201, 150, 105])))

TESTS.append(("Flipd 3d", "3D", 0, True, Flipd(KEYS, [1, 2])))
TESTS.append(("Flipd 3d", "3D", 0, True, Flipd(KEYS, [1, 2])))

TESTS.append(("RandFlipd 3d", "3D", 0, True, RandFlipd(KEYS, 1, [1, 2])))

TESTS.append(("RandAxisFlipd 3d", "3D", 0, True, RandAxisFlipd(KEYS, 1)))
TESTS.append(("RandAxisFlipd 3d", "3D", 0, True, RandAxisFlipd(KEYS, 1)))

for acc in [True, False]:
    TESTS.append(("Orientationd 3d", "3D", 0, True, Orientationd(KEYS, "RAS", as_closest_canonical=acc)))

TESTS.append(("Rotate90d 2d", "2D", 0, True, Rotate90d(KEYS)))

TESTS.append(("Rotate90d 3d", "3D", 0, True, Rotate90d(KEYS, k=2, spatial_axes=(1, 2))))

TESTS.append(("RandRotate90d 3d", "3D", 0, True, RandRotate90d(KEYS, prob=1, spatial_axes=(1, 2))))

TESTS.append(("Spacingd 3d", "3D", 3e-2, True, Spacingd(KEYS, [0.5, 0.7, 0.9], diagonal=False)))

TESTS.append(("Resized 2d", "2D", 2e-1, True, Resized(KEYS, [50, 47])))

TESTS.append(("Resized 3d", "3D", 5e-2, True, Resized(KEYS, [201, 150, 78])))

TESTS.append(("Resized longest 2d", "2D", 2e-1, True, Resized(KEYS, 47, "longest", "area")))

TESTS.append(("Resized longest 3d", "3D", 5e-2, True, Resized(KEYS, 201, "longest", "trilinear", True)))

TESTS.append(
    ("Lambdad 2d", "2D", 5e-2, False, Lambdad(KEYS, func=lambda x: x + 5, inv_func=lambda x: x - 5, overwrite=True))
)

TESTS.append(
    (
        "RandLambdad 3d",
        "3D",
        5e-2,
        False,
        RandLambdad(KEYS, func=lambda x: x * 10, inv_func=lambda x: x / 10, overwrite=True, prob=0.5),
    )
)

TESTS.append(("Zoomd 1d", "1D odd", 0, True, Zoomd(KEYS, zoom=2, keep_size=False)))

TESTS.append(("Zoomd 2d", "2D", 2e-1, True, Zoomd(KEYS, zoom=0.9)))

TESTS.append(("Zoomd 3d", "3D", 3e-2, True, Zoomd(KEYS, zoom=[2.5, 1, 3], keep_size=False)))

TESTS.append(("RandZoom 3d", "3D", 9e-2, True, RandZoomd(KEYS, 1, [0.5, 0.6, 0.9], [1.1, 1, 1.05], keep_size=True)))

TESTS.append(("RandRotated, prob 0", "2D", 0, True, RandRotated(KEYS, prob=0, dtype=np.float64)))

TESTS.append(
    (
        "Rotated 2d",
        "2D",
        8e-2,
        True,
        Rotated(KEYS, random.uniform(np.pi / 6, np.pi), keep_size=True, align_corners=False, dtype=np.float64),
    )
)

TESTS.append(
    (
        "Rotated 3d",
        "3D",
        1e-1,
        True,
        Rotated(KEYS, [random.uniform(np.pi / 6, np.pi) for _ in range(3)], True, dtype=np.float64),
    )
)

TESTS.append(
    (
        "RandRotated 3d",
        "3D",
        1e-1,
        True,
        RandRotated(KEYS, *[random.uniform(np.pi / 6, np.pi) for _ in range(3)], 1, dtype=np.float64),  # type: ignore
    )
)

TESTS.append(("Transposed 2d", "2D", 0, False, Transposed(KEYS, [0, 2, 1])))  # channel=0

TESTS.append(("Transposed 3d", "3D", 0, False, Transposed(KEYS, [0, 3, 1, 2])))  # channel=0

TESTS.append(
    (
        "Affine 3d",
        "3D",
        1e-1,
        True,
        Affined(
            KEYS,
            spatial_size=[155, 179, 192],
            rotate_params=[np.pi / 6, -np.pi / 5, np.pi / 7],
            shear_params=[0.5, 0.5],
            translate_params=[10, 5, -4],
            scale_params=[0.8, 1.3],
        ),
    )
)

TESTS.append(
    (
        "RandAffine 3d",
        "3D",
        1e-1,
        True,
        RandAffined(
            KEYS,
            [155, 179, 192],
            prob=1,
            padding_mode="zeros",
            rotate_range=[np.pi / 6, -np.pi / 5, np.pi / 7],
            shear_range=[(0.5, 0.5)],
            translate_range=[10, 5, -4],
            scale_range=[(0.8, 1.2), (0.9, 1.3)],
        ),
    )
)

TESTS.append(("RandAffine 3d", "3D", 0, True, RandAffined(KEYS, spatial_size=None, prob=0)))

TESTS.append(
    (
        "RandCropByLabelClassesd 2d",
        "2D",
        1e-7,
        True,
        RandCropByLabelClassesd(KEYS, "label", (99, 96), ratios=[1, 2, 3, 4, 5], num_classes=5, num_samples=10),
    )
)

TESTS.append(
    ("RandCropByPosNegLabeld 2d", "2D", 1e-7, True, RandCropByPosNegLabeld(KEYS, "label", (99, 96), num_samples=10))
)

TESTS.append(("RandSpatialCropSamplesd 2d", "2D", 1e-7, True, RandSpatialCropSamplesd(KEYS, (90, 91), num_samples=10)))

TESTS.append(("RandWeightedCropd 2d", "2D", 1e-7, True, RandWeightedCropd(KEYS, "label", (90, 91), num_samples=10)))

TESTS_COMPOSE_X2 = [(t[0] + " Compose", t[1], t[2], t[3], Compose(Compose(t[4:]))) for t in TESTS]

TESTS = TESTS + TESTS_COMPOSE_X2

NUM_SAMPLES = 5
N_SAMPLES_TESTS = [
    [RandCropByLabelClassesd(KEYS, "label", (110, 99), [1, 2, 3, 4, 5], num_classes=5, num_samples=NUM_SAMPLES)],
    [RandCropByPosNegLabeld(KEYS, "label", (110, 99), num_samples=NUM_SAMPLES)],
    [RandSpatialCropSamplesd(KEYS, (90, 91), num_samples=NUM_SAMPLES, random_size=False)],
    [RandWeightedCropd(KEYS, "label", (90, 91), num_samples=NUM_SAMPLES)],
]


def no_collation(x):
    return x


class TestInverse(unittest.TestCase):
    """Test inverse methods.

    If tests are failing, the following function might be useful for displaying
    `x`, `fx`, `f⁻¹fx` and `x - f⁻¹fx`.

    .. code-block:: python

        def plot_im(orig, fwd_bck, fwd):
            import matplotlib.pyplot as plt
            diff_orig_fwd_bck = orig - fwd_bck
            ims_to_show = [orig, fwd, fwd_bck, diff_orig_fwd_bck]
            titles = ["x", "fx", "f⁻¹fx", "x - f⁻¹fx"]
            fig, axes = plt.subplots(1, 4, gridspec_kw={"width_ratios": [i.shape[1] for i in ims_to_show]})
            vmin = min(np.array(i).min() for i in [orig, fwd_bck, fwd])
            vmax = max(np.array(i).max() for i in [orig, fwd_bck, fwd])
            for im, title, ax in zip(ims_to_show, titles, axes):
                _vmin, _vmax = (vmin, vmax) if id(im) != id(diff_orig_fwd_bck) else (None, None)
                im = np.squeeze(np.array(im))
                while im.ndim > 2:
                    im = im[..., im.shape[-1] // 2]
                im_show = ax.imshow(np.squeeze(im), vmin=_vmin, vmax=_vmax)
                ax.set_title(title, fontsize=25)
                ax.axis("off")
                fig.colorbar(im_show, ax=ax)
            plt.show()

    This can then be added to the exception:

    .. code-block:: python

        except AssertionError:
            print(
                f"Failed: {name}. Mean diff = {mean_diff} (expected <= {acceptable_diff}), unmodified diff: {unmodded_diff}"
            )
            if orig[0].ndim > 1:
                plot_im(orig, fwd_bck, unmodified)
    """

    def setUp(self):
        if not has_nib:
            self.skipTest("nibabel required for test_inverse")

        set_determinism(seed=0)

        self.all_data = {}

        affine = make_rand_affine()
        affine[0] *= 2

        for size in [10, 11]:
            # pad 5 onto both ends so that cropping can be lossless
            im_1d = np.pad(np.arange(size), 5)[None]
            name = "1D even" if size % 2 == 0 else "1D odd"
            self.all_data[name] = {
                "image": torch.as_tensor(np.array(im_1d, copy=True)),
                "label": torch.as_tensor(np.array(im_1d, copy=True)),
                "other": torch.as_tensor(np.array(im_1d, copy=True)),
            }

        im_2d_fname, seg_2d_fname = (make_nifti_image(i) for i in create_test_image_2d(101, 100))
        im_3d_fname, seg_3d_fname = (make_nifti_image(i, affine) for i in create_test_image_3d(100, 101, 107))

        load_ims = Compose(
            [LoadImaged(KEYS), EnsureChannelFirstd(KEYS, channel_dim="no_channel"), FromMetaTensord(KEYS)]
        )
        self.all_data["2D"] = load_ims({"image": im_2d_fname, "label": seg_2d_fname})
        self.all_data["3D"] = load_ims({"image": im_3d_fname, "label": seg_3d_fname})

    def tearDown(self):
        set_determinism(seed=None)

    def check_inverse(self, name, keys, orig_d, fwd_bck_d, unmodified_d, acceptable_diff):
        for key in keys:
            orig = orig_d[key]
            fwd_bck = fwd_bck_d[key]
            if isinstance(fwd_bck, torch.Tensor):
                fwd_bck = fwd_bck.cpu().numpy()
            unmodified = unmodified_d[key]
            if isinstance(orig, np.ndarray):
                mean_diff = np.mean(np.abs(orig - fwd_bck))
                resized = ResizeWithPadOrCrop(orig.shape[1:])(unmodified)
                if isinstance(resized, torch.Tensor):
                    resized = resized.detach().cpu().numpy()
                unmodded_diff = np.mean(np.abs(orig - resized))
                try:
                    self.assertLessEqual(mean_diff, acceptable_diff)
                except AssertionError:
                    print(
                        f"Failed: {name}. Mean diff = {mean_diff} (expected <= {acceptable_diff}), unmodified diff: {unmodded_diff}"
                    )
                    if orig[0].ndim == 1:
                        print("orig", orig[0])
                        print("fwd_bck", fwd_bck[0])
                        print("unmod", unmodified[0])
                    raise

    @parameterized.expand(TESTS)
    def test_inverse(self, _, data_name, acceptable_diff, is_meta, *transforms):
        name = _

        data = self.all_data[data_name]
        if is_meta:
            data = ToMetaTensord(KEYS)(data)

        forwards = [data.copy()]

        # Apply forwards
        for t in transforms:
            if isinstance(t, Randomizable):
                t.set_random_state(seed=get_seed())
            forwards.append(t(forwards[-1]))

        # Apply inverses
        fwd_bck = forwards[-1].copy()
        for i, t in enumerate(reversed(transforms)):
            if isinstance(t, InvertibleTransform):
                if isinstance(fwd_bck, list):
                    for j, _fwd_bck in enumerate(fwd_bck):
                        fwd_bck = t.inverse(_fwd_bck)
                        self.check_inverse(
                            name, data.keys(), forwards[-i - 2], fwd_bck, forwards[-1][j], acceptable_diff
                        )
                else:
                    fwd_bck = t.inverse(fwd_bck)
                    self.check_inverse(name, data.keys(), forwards[-i - 2], fwd_bck, forwards[-1], acceptable_diff)

    # skip this test if multiprocessing uses 'spawn', as the check is only basic anyway
    @skipUnless(torch.multiprocessing.get_start_method() == "spawn", "requires spawn")
    def test_fail(self):
        t1 = SpatialPadd("image", [10, 5])
        data = t1(self.all_data["2D"])

        # Check that error is thrown when inverse are used out of order.
        t2 = ResizeWithPadOrCropd("image", [10, 5])
        with self.assertRaises(RuntimeError):
            t2.inverse(data)

    @parameterized.expand(N_SAMPLES_TESTS)
    def test_inverse_inferred_seg(self, extra_transform):
        test_data = []
        for _ in range(20):
            image, label = create_test_image_2d(100, 101)
            test_data.append({"image": image, "label": label.astype(np.float32)})

        batch_size = 10
        # num workers = 0 for mac
        num_workers = 2 if sys.platform == "linux" else 0
        transforms = Compose(
            [EnsureChannelFirstd(KEYS, channel_dim="no_channel"), SpatialPadd(KEYS, (150, 153)), extra_transform]
        )

        dataset = CacheDataset(test_data, transform=transforms, progress=False)
        loader = DataLoader(dataset, batch_size=batch_size, shuffle=False, num_workers=num_workers)

        device = "cuda" if torch.cuda.is_available() else "cpu"
        model = UNet(spatial_dims=2, in_channels=1, out_channels=1, channels=(2, 4), strides=(1,)).to(device)

        data = first(loader)
        self.assertEqual(data["image"].shape[0], batch_size * NUM_SAMPLES)

        labels = data["label"].to(device)
        self.assertIsInstance(labels, MetaTensor)
        segs = model(labels).detach().cpu()
        segs_decollated = decollate_batch(segs)
        self.assertIsInstance(segs_decollated[0], MetaTensor)
        # inverse of individual segmentation
        seg_metatensor = first(segs_decollated)
        # test to convert interpolation mode for 1 data of model output batch
        convert_applied_interp_mode(seg_metatensor.applied_operations, mode="nearest", align_corners=None)

        # manually invert the last crop samples
        xform = seg_metatensor.applied_operations.pop(-1)
        shape_before_extra_xform = xform["orig_size"]
        resizer = ResizeWithPadOrCrop(spatial_size=shape_before_extra_xform)
        with resizer.trace_transform(False):
            seg_metatensor = resizer(seg_metatensor)
        no_ops_id_tensor = reset_ops_id(deepcopy(seg_metatensor))

        with allow_missing_keys_mode(transforms):
            inv_seg = transforms.inverse({"label": seg_metatensor})["label"]
            inv_seg_1 = transforms.inverse({"label": no_ops_id_tensor})["label"]
        self.assertEqual(inv_seg.shape[1:], test_data[0]["label"].shape)
        self.assertEqual(inv_seg_1.shape[1:], test_data[0]["label"].shape)

        # # Inverse of batch
        # batch_inverter = BatchInverseTransform(transforms, loader, collate_fn=no_collation, detach=True)
        # with allow_missing_keys_mode(transforms):
        #     inv_batch = batch_inverter(first(loader))
        # self.assertEqual(inv_batch[0]["label"].shape[1:], test_data[0]["label"].shape)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.metrics import GeneralizedDiceScore, compute_generalized_dice

_device = "cuda:0" if torch.cuda.is_available() else "cpu"

# keep background
TEST_CASE_1 = [  # y (1, 1, 2, 2), y_pred (1, 1, 2, 2), expected out (1)
    {
        "y_pred": torch.tensor([[[[1.0, 0.0], [0.0, 1.0]]]], device=_device),
        "y": torch.tensor([[[[1.0, 0.0], [1.0, 1.0]]]], device=_device),
        "include_background": True,
    },
    [0.8],
]

# remove background
TEST_CASE_2 = [  # y (2, 1, 2, 2), y_pred (2, 3, 2, 2), expected out (2) (no background)
    {
        "y_pred": torch.tensor(
            [
                [[[0.0, 1.0], [0.0, 0.0]], [[0.0, 0.0], [1.0, 1.0]], [[1.0, 0.0], [0.0, 0.0]]],
                [[[0.0, 0.0], [0.0, 1.0]], [[1.0, 0.0], [0.0, 0.0]], [[0.0, 1.0], [1.0, 0.0]]],
            ]
        ),
        "y": torch.tensor(
            [
                [[[0.0, 0.0], [0.0, 1.0]], [[1.0, 0.0], [1.0, 0.0]], [[0.0, 1.0], [0.0, 0.0]]],
                [[[0.0, 0.0], [0.0, 1.0]], [[1.0, 1.0], [0.0, 0.0]], [[0.0, 0.0], [1.0, 0.0]]],
            ]
        ),
        "include_background": False,
    },
    [0.1667, 0.6667],
]

# should return 0 for both cases
TEST_CASE_3 = [
    {
        "y_pred": torch.tensor(
            [
                [[[0.0, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]], [[1.0, 1.0], [1.0, 1.0]]],
                [[[0.0, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]], [[1.0, 1.0], [1.0, 1.0]]],
            ]
        ),
        "y": torch.tensor(
            [
                [[[1.0, 1.0], [1.0, 1.0]], [[0.0, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]]],
                [[[0.0, 1.0], [1.0, 0.0]], [[1.0, 0.0], [0.0, 1.0]], [[0.0, 0.0], [0.0, 0.0]]],
            ]
        ),
        "include_background": True,
    },
    [0.0, 0.0],
]

TEST_CASE_4 = [
    {"include_background": True, "reduction": "mean_batch"},
    {
        "y_pred": torch.tensor(
            [
                [[[1.0, 1.0], [1.0, 0.0]], [[0.0, 1.0], [0.0, 0.0]], [[0.0, 1.0], [1.0, 1.0]]],
                [[[1.0, 0.0], [1.0, 1.0]], [[0.0, 1.0], [1.0, 1.0]], [[0.0, 1.0], [1.0, 0.0]]],
            ]
        ),
        "y": torch.tensor(
            [
                [[[1.0, 1.0], [1.0, 1.0]], [[0.0, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]]],
                [[[0.0, 0.0], [0.0, 1.0]], [[1.0, 1.0], [0.0, 0.0]], [[0.0, 0.0], [1.0, 0.0]]],
            ]
        ),
    },
    [0.5455],
]

TEST_CASE_5 = [
    {"include_background": True, "reduction": "sum_batch"},
    {
        "y_pred": torch.tensor(
            [
                [[[1.0, 1.0], [1.0, 0.0]], [[0.0, 1.0], [0.0, 0.0]], [[0.0, 1.0], [1.0, 1.0]]],
                [[[1.0, 0.0], [1.0, 1.0]], [[0.0, 1.0], [1.0, 1.0]], [[0.0, 1.0], [1.0, 0.0]]],
            ]
        ),
        "y": torch.tensor(
            [
                [[[1.0, 1.0], [1.0, 1.0]], [[0.0, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]]],
                [[[1.0, 1.0], [1.0, 1.0]], [[0.0, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]]],
            ]
        ),
    },
    1.0455,
]

TEST_CASE_6 = [{"y": torch.ones((2, 2, 3, 3)), "y_pred": torch.ones((2, 2, 3, 3))}, [1.0000, 1.0000]]

TEST_CASE_7 = [{"y": torch.zeros((2, 2, 3, 3)), "y_pred": torch.ones((2, 2, 3, 3))}, [0.0000, 0.0000]]

TEST_CASE_8 = [{"y": torch.ones((2, 2, 3, 3)), "y_pred": torch.zeros((2, 2, 3, 3))}, [0.0000, 0.0000]]

TEST_CASE_9 = [{"y": torch.zeros((2, 2, 3, 3)), "y_pred": torch.zeros((2, 2, 3, 3))}, [1.0000, 1.0000]]


class TestComputeGeneralizedDiceScore(unittest.TestCase):

    @parameterized.expand([TEST_CASE_1])
    def test_device(self, input_data, _expected_value):
        result = compute_generalized_dice(**input_data)
        np.testing.assert_equal(result.device, input_data["y_pred"].device)

    # Functional part tests
    @parameterized.expand([TEST_CASE_1, TEST_CASE_2, TEST_CASE_6, TEST_CASE_7, TEST_CASE_8, TEST_CASE_9])
    def test_value(self, input_data, expected_value):
        result = compute_generalized_dice(**input_data)
        np.testing.assert_allclose(result.cpu().numpy(), expected_value, atol=1e-4)

    # Functional part tests
    @parameterized.expand([TEST_CASE_3])
    def test_nans(self, input_data, expected_value):
        result = compute_generalized_dice(**input_data)
        self.assertTrue(np.allclose(np.isnan(result.cpu().numpy()), expected_value))

    # Samplewise tests
    @parameterized.expand([TEST_CASE_1, TEST_CASE_2])
    def test_value_class(self, input_data, expected_value):
        # same test as for compute_meandice
        vals = {}
        vals["y_pred"] = input_data.pop("y_pred")
        vals["y"] = input_data.pop("y")
        generalized_dice_score = GeneralizedDiceScore(**input_data)
        generalized_dice_score(**vals)
        result = generalized_dice_score.aggregate(reduction="none")
        np.testing.assert_allclose(result.cpu().numpy(), expected_value, atol=1e-4)

    # Aggregation tests
    @parameterized.expand([TEST_CASE_4, TEST_CASE_5])
    def test_nans_class(self, params, input_data, expected_value):
        generalized_dice_score = GeneralizedDiceScore(**params)
        generalized_dice_score(**input_data)
        result = generalized_dice_score.aggregate()
        np.testing.assert_allclose(result.cpu().numpy(), expected_value, atol=1e-4)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.networks import eval_mode
from monai.networks.blocks.mlp import MLPBlock

TEST_CASE_MLP = []
for dropout_rate in np.linspace(0, 1, 4):
    for hidden_size in [128, 256, 512, 768]:
        for mlp_dim in [0, 1028, 2048, 3072]:
            test_case = [
                {"hidden_size": hidden_size, "mlp_dim": mlp_dim, "dropout_rate": dropout_rate},
                (2, 512, hidden_size),
                (2, 512, hidden_size),
            ]
            TEST_CASE_MLP.append(test_case)


class TestMLPBlock(unittest.TestCase):

    @parameterized.expand(TEST_CASE_MLP)
    def test_shape(self, input_param, input_shape, expected_shape):
        net = MLPBlock(**input_param)
        with eval_mode(net):
            result = net(torch.randn(input_shape))
            self.assertEqual(result.shape, expected_shape)

    def test_ill_arg(self):
        with self.assertRaises(ValueError):
            MLPBlock(hidden_size=128, mlp_dim=512, dropout_rate=5.0)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest
from copy import deepcopy

import numpy as np
import torch
from parameterized import parameterized

from monai.data import MetaTensor, set_track_meta
from monai.transforms import Affine, Resize
from monai.transforms.lazy.functional import apply_pending
from monai.utils import optional_import
from tests.lazy_transforms_utils import test_resampler_lazy
from tests.utils import TEST_NDARRAYS_ALL, assert_allclose, test_local_inversion

TESTS = []
for p in TEST_NDARRAYS_ALL:
    for device in [None, "cpu", "cuda"] if torch.cuda.is_available() else [None, "cpu"]:
        TESTS.append(
            [
                dict(padding_mode="zeros", device=device),
                {"img": p(np.arange(9).reshape((1, 3, 3))), "spatial_size": (-1, 0)},
                p(np.arange(9).reshape(1, 3, 3)),
            ]
        )
        TESTS.append(
            [
                dict(padding_mode="zeros", device=device, image_only=True),
                {"img": p(np.arange(9).reshape((1, 3, 3))), "spatial_size": (-1, 0)},
                p(np.arange(9).reshape(1, 3, 3)),
            ]
        )
        TESTS.append(
            [
                dict(padding_mode="zeros", device=device),
                {"img": p(np.arange(4).reshape((1, 2, 2)))},
                p(np.arange(4).reshape(1, 2, 2)),
            ]
        )
        TESTS.append(
            [
                dict(padding_mode="zeros", device=device),
                {"img": p(np.arange(4).reshape((1, 2, 2))), "spatial_size": (4, 4)},
                p(np.array([[[0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0], [0.0, 2.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]])),
            ]
        )
        TESTS.append(
            [
                dict(rotate_params=[np.pi / 2], padding_mode="zeros", device=device),
                {"img": p(np.arange(4).reshape((1, 2, 2))), "spatial_size": (4, 4)},
                p(np.array([[[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 0.0, 0.0], [0.0, 3.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]])),
            ]
        )
        TESTS.append(
            [
                dict(rotate_params=[np.pi / 2], padding_mode="zeros", device=device, align_corners=False),
                {"img": p(np.arange(4).reshape((1, 2, 2))), "spatial_size": (4, 4)},
                p(np.array([[[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 0.0, 0.0], [0.0, 3.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]])),
            ]
        )
        TESTS.append(
            [
                dict(
                    affine=p(torch.tensor([[0.0, -1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]])),
                    padding_mode="zeros",
                    device=device,
                ),
                {"img": p(np.arange(4).reshape((1, 2, 2))), "spatial_size": (4, 4)},
                p(np.array([[[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 0.0, 0.0], [0.0, 3.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]])),
            ]
        )
        TESTS.append(
            [
                dict(padding_mode="zeros", device=device),
                {"img": p(np.arange(27).reshape((1, 3, 3, 3))), "spatial_size": (-1, 0, 0)},
                p(np.arange(27).reshape(1, 3, 3, 3)),
            ]
        )
        TESTS.append(
            [
                dict(padding_mode="zeros", device=device),
                {"img": p(np.arange(8).reshape((1, 2, 2, 2))), "spatial_size": (4, 4, 4)},
                p(
                    np.array(
                        [
                            [
                                [
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                ],
                                [
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 1.0, 0.0],
                                    [0.0, 2.0, 3.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                ],
                                [
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 4.0, 5.0, 0.0],
                                    [0.0, 6.0, 7.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                ],
                                [
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                ],
                            ]
                        ]
                    )
                ),
            ]
        )
        TESTS.append(
            [
                dict(rotate_params=[np.pi / 2], padding_mode="zeros", device=device),
                {"img": p(np.arange(8).reshape((1, 2, 2, 2))), "spatial_size": (4, 4, 4)},
                p(
                    np.array(
                        [
                            [
                                [
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                ],
                                [
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 2.0, 0.0, 0.0],
                                    [0.0, 3.0, 1.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                ],
                                [
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 6.0, 4.0, 0.0],
                                    [0.0, 7.0, 5.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                ],
                                [
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                    [0.0, 0.0, 0.0, 0.0],
                                ],
                            ]
                        ]
                    )
                ),
            ]
        )


class TestAffine(unittest.TestCase):

    @parameterized.expand(TESTS)
    def test_affine(self, input_param, input_data, expected_val):
        input_copy = deepcopy(input_data["img"])
        g = Affine(**input_param)
        result = g(**input_data)
        output_idx = None
        if isinstance(result, tuple):
            output_idx = 0
            result = result[output_idx]

        test_local_inversion(g, result, input_copy)
        assert_allclose(result, expected_val, rtol=1e-4, atol=1e-4, type_test=False)

        set_track_meta(False)
        result = g(**input_data)
        if isinstance(result, tuple):
            result = result[0]
        self.assertNotIsInstance(result, MetaTensor)
        self.assertIsInstance(result, torch.Tensor)
        set_track_meta(True)

        # test lazy
        lazy_input_param = input_param.copy()
        for align_corners in [True, False]:
            lazy_input_param["align_corners"] = align_corners
            resampler = Affine(**lazy_input_param)
            non_lazy_result = resampler(**input_data)
            test_resampler_lazy(resampler, non_lazy_result, lazy_input_param, input_data, output_idx=output_idx)


@unittest.skipUnless(optional_import("scipy")[1], "Requires scipy library.")
class TestAffineConsistency(unittest.TestCase):

    @parameterized.expand([[7], [8], [9]])
    def test_affine_resize(self, s):
        """s"""
        im = np.arange(4).reshape(1, 2, 2).astype(float)
        mat = np.array([[1 / s, 0, 0], [0, 1 / s, 0], [0, 0, 1]])
        sp_size = 2 * s

        def method_0(im, ac):
            xform = Affine(align_corners=ac, affine=mat, image_only=True, spatial_size=sp_size)
            xform.lazy = True
            out = xform(im)
            overrides = {"padding_mode": "border", "align_corners": ac}
            out = apply_pending(out, overrides=overrides)[0]
            return out

        def method_1(im, ac):
            xform = Affine(align_corners=ac, affine=mat, image_only=True, spatial_size=sp_size)
            xform.lazy = True
            out = xform(im)
            overrides = {"mode": 1, "padding_mode": "nearest", "align_corners": ac}
            out = apply_pending(out, overrides=overrides)[0]
            return out

        def method_2(im, ac):
            xform = Affine(align_corners=ac, affine=mat, padding_mode="border", image_only=True, spatial_size=sp_size)
            out = xform(im)
            return out

        def method_3(im, ac):
            xform = Affine(
                align_corners=ac, affine=mat, mode=1, padding_mode="nearest", image_only=True, spatial_size=sp_size
            )
            out = xform(im)
            return out

        for call in (method_0, method_1, method_2, method_3):
            for ac in (False, True):
                out = call(im, ac)
                ref = Resize(align_corners=ac, spatial_size=(sp_size, sp_size), mode="bilinear")(im)
                assert_allclose(out, ref, rtol=1e-4, atol=1e-4, type_test=False)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
import torch

from monai.config import NdarrayTensor
from monai.transforms import rescale_array
from monai.utils import convert_data_type, optional_import

PIL, _ = optional_import("PIL")
GifImage, _ = optional_import("PIL.GifImagePlugin", name="Image")

if TYPE_CHECKING:
    from tensorboard.compat.proto.summary_pb2 import Summary
    from tensorboardX import SummaryWriter as SummaryWriterX
    from tensorboardX.proto.summary_pb2 import Summary as SummaryX
    from torch.utils.tensorboard import SummaryWriter

    has_tensorboardx = True
else:
    Summary, _ = optional_import("tensorboard.compat.proto.summary_pb2", name="Summary")
    SummaryX, _ = optional_import("tensorboardX.proto.summary_pb2", name="Summary")
    SummaryWriter, _ = optional_import("torch.utils.tensorboard", name="SummaryWriter")
    SummaryWriterX, has_tensorboardx = optional_import("tensorboardX", name="SummaryWriter")

__all__ = ["make_animated_gif_summary", "add_animated_gif", "plot_2d_or_3d_image"]


def _image3_animated_gif(
    tag: str,
    image: np.ndarray | torch.Tensor,
    writer: SummaryWriter | SummaryWriterX | None,
    frame_dim: int = 0,
    scale_factor: float = 1.0,
) -> Any:
    """Function to actually create the animated gif.

    Args:
        tag: Data identifier
        image: 3D image tensors expected to be in `HWD` format
        writer: the tensorboard writer to plot image
        frame_dim: the dimension used as frames for GIF image, expect data shape as `HWD`, default to `0`.
        scale_factor: amount to multiply values by. if the image data is between 0 and 1, using 255 for this value will
            scale it to displayable range
    """
    if len(image.shape) != 3:
        raise AssertionError("3D image tensors expected to be in `HWD` format, len(image.shape) != 3")

    image_np, *_ = convert_data_type(image, output_type=np.ndarray)
    ims = [(i * scale_factor).astype(np.uint8, copy=False) for i in np.moveaxis(image_np, frame_dim, 0)]
    ims = [GifImage.fromarray(im) for im in ims]
    img_str = b""
    for b_data in PIL.GifImagePlugin.getheader(ims[0])[0]:
        img_str += b_data
    img_str += b"\x21\xFF\x0B\x4E\x45\x54\x53\x43\x41\x50" b"\x45\x32\x2E\x30\x03\x01\x00\x00\x00"
    for i in ims:
        for b_data in PIL.GifImagePlugin.getdata(i):
            img_str += b_data
    img_str += b"\x3B"

    summary = SummaryX if has_tensorboardx and isinstance(writer, SummaryWriterX) else Summary
    summary_image_str = summary.Image(height=10, width=10, colorspace=1, encoded_image_string=img_str)
    image_summary = summary.Value(tag=tag, image=summary_image_str)
    return summary(value=[image_summary])


def make_animated_gif_summary(
    tag: str,
    image: np.ndarray | torch.Tensor,
    writer: SummaryWriter | SummaryWriterX | None = None,
    max_out: int = 3,
    frame_dim: int = -3,
    scale_factor: float = 1.0,
) -> Summary:
    """Creates an animated gif out of an image tensor in 'CHWD' format and returns Summary.

    Args:
        tag: Data identifier
        image: The image, expected to be in `CHWD` format
        writer: the tensorboard writer to plot image
        max_out: maximum number of image channels to animate through
        frame_dim: the dimension used as frames for GIF image, expect input data shape as `CHWD`,
            default to `-3` (the first spatial dim)
        scale_factor: amount to multiply values by.
            if the image data is between 0 and 1, using 255 for this value will scale it to displayable range
    """

    suffix = "/image" if max_out == 1 else "/image/{}"
    # GIF image has no channel dim, reduce the spatial dim index if positive
    frame_dim = frame_dim - 1 if frame_dim > 0 else frame_dim

    summary_op = []
    for it_i in range(min(max_out, list(image.shape)[0])):
        one_channel_img: torch.Tensor | np.ndarray = (
            image[it_i, :, :, :].squeeze(dim=0) if isinstance(image, torch.Tensor) else image[it_i, :, :, :]
        )
        summary_op.append(
            _image3_animated_gif(tag + suffix.format(it_i), one_channel_img, writer, frame_dim, scale_factor)
        )
    return summary_op


def add_animated_gif(
    writer: SummaryWriter | SummaryWriterX,
    tag: str,
    image_tensor: np.ndarray | torch.Tensor,
    max_out: int = 3,
    frame_dim: int = -3,
    scale_factor: float = 1.0,
    global_step: int | None = None,
) -> None:
    """Creates an animated gif out of an image tensor in 'CHWD' format and writes it with SummaryWriter.

    Args:
        writer: Tensorboard SummaryWriter to write to
        tag: Data identifier
        image_tensor: tensor for the image to add, expected to be in `CHWD` format
        max_out: maximum number of image channels to animate through
        frame_dim: the dimension used as frames for GIF image, expect input data shape as `CHWD`,
            default to `-3` (the first spatial dim)
        scale_factor: amount to multiply values by. If the image data is between 0 and 1, using 255 for this value will
            scale it to displayable range
        global_step: Global step value to record
    """
    summary = make_animated_gif_summary(
        tag=tag, image=image_tensor, writer=writer, max_out=max_out, frame_dim=frame_dim, scale_factor=scale_factor
    )
    for s in summary:
        # add GIF for every channel separately
        writer._get_file_writer().add_summary(s, global_step)


def plot_2d_or_3d_image(
    data: NdarrayTensor | list[NdarrayTensor],
    step: int,
    writer: SummaryWriter | SummaryWriterX,
    index: int = 0,
    max_channels: int = 1,
    frame_dim: int = -3,
    max_frames: int = 24,
    tag: str = "output",
) -> None:
    """Plot 2D or 3D image on the TensorBoard, 3D image will be converted to GIF image.

    Note:
        Plot 3D or 2D image(with more than 3 channels) as separate images.
        And if writer is from TensorBoardX, data has 3 channels and `max_channels=3`, will plot as RGB video.

    Args:
        data: target data to be plotted as image on the TensorBoard.
            The data is expected to have 'NCHW[D]' dimensions or a list of data with `CHW[D]` dimensions,
            and only plot the first in the batch.
        step: current step to plot in a chart.
        writer: specify TensorBoard or TensorBoardX SummaryWriter to plot the image.
        index: plot which element in the input data batch, default is the first element.
        max_channels: number of channels to plot.
        frame_dim: if plotting 3D image as GIF, specify the dimension used as frames,
            expect input data shape as `NCHWD`, default to `-3` (the first spatial dim)
        max_frames: if plot 3D RGB image as video in TensorBoardX, set the FPS to `max_frames`.
        tag: tag of the plotted image on TensorBoard.
    """
    data_index = data[index]
    # as the `d` data has no batch dim, reduce the spatial dim index if positive
    frame_dim = frame_dim - 1 if frame_dim > 0 else frame_dim

    d: np.ndarray = data_index.detach().cpu().numpy() if isinstance(data_index, torch.Tensor) else data_index

    if d.ndim == 2:
        d = rescale_array(d, 0, 1)  # type: ignore
        dataformats = "HW"
        writer.add_image(f"{tag}_{dataformats}", d, step, dataformats=dataformats)
        return

    if d.ndim == 3:
        if d.shape[0] == 3 and max_channels == 3:  # RGB
            dataformats = "CHW"
            writer.add_image(f"{tag}_{dataformats}", d, step, dataformats=dataformats)
            return
        dataformats = "HW"
        for j, d2 in enumerate(d[:max_channels]):
            d2 = rescale_array(d2, 0, 1)
            writer.add_image(f"{tag}_{dataformats}_{j}", d2, step, dataformats=dataformats)
        return

    if d.ndim >= 4:
        spatial = d.shape[-3:]
        d = d.reshape([-1] + list(spatial))
        if d.shape[0] == 3 and max_channels == 3 and has_tensorboardx and isinstance(writer, SummaryWriterX):  # RGB
            # move the expected frame dim to the end as `T` dim for video
            d = np.moveaxis(d, frame_dim, -1)
            writer.add_video(tag, d[None], step, fps=max_frames, dataformats="NCHWT")
            return
        # scale data to 0 - 255 for visualization
        max_channels = min(max_channels, d.shape[0])
        d = np.stack([rescale_array(i, 0, 255) for i in d[:max_channels]], axis=0)
        # will plot every channel as a separate GIF image
        add_animated_gif(writer, f"{tag}_HWD", d, max_out=max_channels, frame_dim=frame_dim, global_step=step)
        return



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks import eval_mode
from monai.networks.nets.vit import ViT
from tests.utils import SkipIfBeforePyTorchVersion, skip_if_quick, test_script_save

TEST_CASE_Vit = []
for dropout_rate in [0.6]:
    for in_channels in [4]:
        for hidden_size in [768]:
            for img_size in [96, 128]:
                for patch_size in [16]:
                    for num_heads in [12]:
                        for mlp_dim in [3072]:
                            for num_layers in [4]:
                                for num_classes in [8]:
                                    for pos_embed in ["conv", "perceptron"]:
                                        for classification in [False, True]:
                                            for nd in (2, 3):
                                                test_case = [
                                                    {
                                                        "in_channels": in_channels,
                                                        "img_size": (img_size,) * nd,
                                                        "patch_size": (patch_size,) * nd,
                                                        "hidden_size": hidden_size,
                                                        "mlp_dim": mlp_dim,
                                                        "num_layers": num_layers,
                                                        "num_heads": num_heads,
                                                        "pos_embed": pos_embed,
                                                        "classification": classification,
                                                        "num_classes": num_classes,
                                                        "dropout_rate": dropout_rate,
                                                    },
                                                    (2, in_channels, *([img_size] * nd)),
                                                    (2, (img_size // patch_size) ** nd, hidden_size),
                                                ]
                                                if nd == 2:
                                                    test_case[0]["spatial_dims"] = 2  # type: ignore
                                                    if classification:
                                                        test_case[0]["post_activation"] = False  # type: ignore
                                                if test_case[0]["classification"]:  # type: ignore
                                                    test_case[2] = (2, test_case[0]["num_classes"])  # type: ignore
                                                TEST_CASE_Vit.append(test_case)


@skip_if_quick
class TestViT(unittest.TestCase):

    @parameterized.expand(TEST_CASE_Vit)
    def test_shape(self, input_param, input_shape, expected_shape):
        net = ViT(**input_param)
        with eval_mode(net):
            result, _ = net(torch.randn(input_shape))
            self.assertEqual(result.shape, expected_shape)

    @parameterized.expand(
        [
            (1, (128, 128, 128), (16, 16, 16), 128, 3072, 12, 12, "conv", False, 5.0),
            (1, (32, 32, 32), (64, 64, 64), 512, 3072, 12, 8, "perceptron", False, 0.3),
            (1, (96, 96, 96), (8, 8, 8), 512, 3072, 12, 14, "conv", False, 0.3),
            (1, (97, 97, 97), (4, 4, 4), 768, 3072, 12, 8, "perceptron", True, 0.3),
            (4, (96, 96, 96), (16, 16, 16), 768, 3072, 12, 12, "perc", False, 0.3),
        ]
    )
    def test_ill_arg(
        self,
        in_channels,
        img_size,
        patch_size,
        hidden_size,
        mlp_dim,
        num_layers,
        num_heads,
        pos_embed,
        classification,
        dropout_rate,
    ):
        with self.assertRaises(ValueError):
            ViT(
                in_channels=in_channels,
                img_size=img_size,
                patch_size=patch_size,
                hidden_size=hidden_size,
                mlp_dim=mlp_dim,
                num_layers=num_layers,
                num_heads=num_heads,
                pos_embed=pos_embed,
                classification=classification,
                dropout_rate=dropout_rate,
            )

    @parameterized.expand(TEST_CASE_Vit)
    @SkipIfBeforePyTorchVersion((1, 9))
    def test_script(self, input_param, input_shape, _):
        net = ViT(**(input_param))
        net.eval()
        with torch.no_grad():
            torch.jit.script(net)

        test_data = torch.randn(input_shape)
        test_script_save(net, test_data)

    def test_access_attn_matrix(self):
        # input format
        in_channels = 1
        img_size = (96, 96, 96)
        patch_size = (16, 16, 16)
        in_shape = (1, in_channels, img_size[0], img_size[1], img_size[2])

        # no data in the matrix
        no_matrix_acess_blk = ViT(in_channels=in_channels, img_size=img_size, patch_size=patch_size)
        no_matrix_acess_blk(torch.randn(in_shape))
        assert isinstance(no_matrix_acess_blk.blocks[0].attn.att_mat, torch.Tensor)
        # no of elements is zero
        assert no_matrix_acess_blk.blocks[0].attn.att_mat.nelement() == 0

        # be able to acess the attention matrix
        matrix_acess_blk = ViT(in_channels=in_channels, img_size=img_size, patch_size=patch_size, save_attn=True)
        matrix_acess_blk(torch.randn(in_shape))
        assert matrix_acess_blk.blocks[0].attn.att_mat.shape == (in_shape[0], 12, 216, 216)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.losses import DiceFocalLoss, DiceLoss, FocalLoss
from tests.utils import test_script_save


class TestDiceFocalLoss(unittest.TestCase):

    def test_result_onehot_target_include_bg(self):
        size = [3, 3, 5, 5]
        label = torch.randint(low=0, high=2, size=size)
        pred = torch.randn(size)
        for reduction in ["sum", "mean", "none"]:
            for weight in [None, torch.tensor([1.0, 1.0, 2.0]), (3, 2.0, 1)]:
                common_params = {
                    "include_background": True,
                    "to_onehot_y": False,
                    "reduction": reduction,
                    "weight": weight,
                }
                for lambda_focal in [0.5, 1.0, 1.5]:
                    dice_focal = DiceFocalLoss(gamma=1.0, lambda_focal=lambda_focal, **common_params)
                    dice = DiceLoss(**common_params)
                    focal = FocalLoss(gamma=1.0, **common_params)
                    result = dice_focal(pred, label)
                    expected_val = dice(pred, label) + lambda_focal * focal(pred, label)
                    np.testing.assert_allclose(result, expected_val)

    @parameterized.expand([[[3, 3, 5, 5], True], [[3, 2, 5, 5], False]])
    def test_result_no_onehot_no_bg(self, size, onehot):
        label = torch.randint(low=0, high=size[1] - 1, size=size)
        if onehot:
            label = torch.argmax(label, dim=1, keepdim=True)
        pred = torch.randn(size)
        for reduction in ["sum", "mean", "none"]:
            for weight in [2.0] + [] if size[1] != 3 else [torch.tensor([1.0, 2.0]), (2.0, 1)]:
                for lambda_focal in [0.5, 1.0, 1.5]:
                    common_params = {
                        "include_background": False,
                        "softmax": True,
                        "to_onehot_y": onehot,
                        "reduction": reduction,
                        "weight": weight,
                    }
                    dice_focal = DiceFocalLoss(lambda_focal=lambda_focal, **common_params)
                    dice = DiceLoss(**common_params)
                    common_params.pop("softmax", None)
                    focal = FocalLoss(**common_params)
                    result = dice_focal(pred, label)
                    expected_val = dice(pred, label) + lambda_focal * focal(pred, label)
                    np.testing.assert_allclose(result, expected_val)

    def test_ill_shape(self):
        loss = DiceFocalLoss()
        with self.assertRaises(AssertionError):
            loss.forward(torch.ones((1, 2, 3)), torch.ones((1, 2, 5)))

    def test_ill_shape2(self):
        loss = DiceFocalLoss()
        with self.assertRaises(ValueError):
            loss.forward(torch.ones((1, 2, 3)), torch.ones((1, 1, 2, 3)))

    def test_ill_shape3(self):
        loss = DiceFocalLoss()
        with self.assertRaises(ValueError):
            loss.forward(torch.ones((1, 3, 4, 4)), torch.ones((1, 2, 4, 4)))

    def test_ill_lambda(self):
        with self.assertRaisesRegex(ValueError, ""):
            DiceFocalLoss(lambda_dice=-1.0)

    def test_script(self):
        loss = DiceFocalLoss()
        test_input = torch.ones(2, 1, 8, 8)
        test_script_save(loss, test_input, test_input)

    @parameterized.expand(
        [
            ("sum_None_0.5_0.25", "sum", None, 0.5, 0.25),
            ("sum_weight_0.5_0.25", "sum", torch.tensor([1.0, 1.0, 2.0]), 0.5, 0.25),
            ("sum_weight_tuple_0.5_0.25", "sum", (3, 2.0, 1), 0.5, 0.25),
            ("mean_None_0.5_0.25", "mean", None, 0.5, 0.25),
            ("mean_weight_0.5_0.25", "mean", torch.tensor([1.0, 1.0, 2.0]), 0.5, 0.25),
            ("mean_weight_tuple_0.5_0.25", "mean", (3, 2.0, 1), 0.5, 0.25),
            ("none_None_0.5_0.25", "none", None, 0.5, 0.25),
            ("none_weight_0.5_0.25", "none", torch.tensor([1.0, 1.0, 2.0]), 0.5, 0.25),
            ("none_weight_tuple_0.5_0.25", "none", (3, 2.0, 1), 0.5, 0.25),
        ]
    )
    def test_with_alpha(self, name, reduction, weight, lambda_focal, alpha):
        size = [3, 3, 5, 5]
        label = torch.randint(low=0, high=2, size=size)
        pred = torch.randn(size)

        common_params = {"include_background": True, "to_onehot_y": False, "reduction": reduction, "weight": weight}

        dice_focal = DiceFocalLoss(gamma=1.0, lambda_focal=lambda_focal, alpha=alpha, **common_params)
        dice = DiceLoss(**common_params)
        focal = FocalLoss(gamma=1.0, alpha=alpha, **common_params)

        result = dice_focal(pred, label)
        expected_val = dice(pred, label) + lambda_focal * focal(pred, label)

        np.testing.assert_allclose(result, expected_val, err_msg=f"Failed on case: {name}")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Sequence, Tuple

import torch
from torch import nn

__all__ = ["VectorQuantizer", "EMAQuantizer"]


class EMAQuantizer(nn.Module):
    """
    Vector Quantization module using Exponential Moving Average (EMA) to learn the codebook parameters based on  Neural
    Discrete Representation Learning by Oord et al. (https://arxiv.org/abs/1711.00937) and the official implementation
    that can be found at https://github.com/deepmind/sonnet/blob/v2/sonnet/src/nets/vqvae.py#L148 and commit
    58d9a2746493717a7c9252938da7efa6006f3739.

    This module is not compatible with TorchScript while working in a Distributed Data Parallelism Module. This is due
    to lack of TorchScript support for torch.distributed module as per https://github.com/pytorch/pytorch/issues/41353
    on 22/10/2022. If you want to TorchScript your model, please turn set `ddp_sync` to False.

    Args:
        spatial_dims: number of spatial dimensions of the input.
        num_embeddings: number of atomic elements in the codebook.
        embedding_dim: number of channels of the input and atomic elements.
        commitment_cost: scaling factor of the MSE loss between input and its quantized version. Defaults to 0.25.
        decay: EMA decay. Defaults to 0.99.
        epsilon: epsilon value. Defaults to 1e-5.
        embedding_init: initialization method for the codebook. Defaults to "normal".
        ddp_sync: whether to synchronize the codebook across processes. Defaults to True.
    """

    def __init__(
        self,
        spatial_dims: int,
        num_embeddings: int,
        embedding_dim: int,
        commitment_cost: float = 0.25,
        decay: float = 0.99,
        epsilon: float = 1e-5,
        embedding_init: str = "normal",
        ddp_sync: bool = True,
    ):
        super().__init__()
        self.spatial_dims: int = spatial_dims
        self.embedding_dim: int = embedding_dim
        self.num_embeddings: int = num_embeddings

        assert self.spatial_dims in [2, 3], ValueError(
            f"EMAQuantizer only supports 4D and 5D tensor inputs but received spatial dims {spatial_dims}."
        )

        self.embedding: torch.nn.Embedding = torch.nn.Embedding(self.num_embeddings, self.embedding_dim)
        if embedding_init == "normal":
            # Initialization is passed since the default one is normal inside the nn.Embedding
            pass
        elif embedding_init == "kaiming_uniform":
            torch.nn.init.kaiming_uniform_(self.embedding.weight.data, mode="fan_in", nonlinearity="linear")
        self.embedding.weight.requires_grad = False

        self.commitment_cost: float = commitment_cost

        self.register_buffer("ema_cluster_size", torch.zeros(self.num_embeddings))
        self.register_buffer("ema_w", self.embedding.weight.data.clone())
        # declare types for mypy
        self.ema_cluster_size: torch.Tensor
        self.ema_w: torch.Tensor
        self.decay: float = decay
        self.epsilon: float = epsilon

        self.ddp_sync: bool = ddp_sync

        # Precalculating required permutation shapes
        self.flatten_permutation = [0] + list(range(2, self.spatial_dims + 2)) + [1]
        self.quantization_permutation: Sequence[int] = [0, self.spatial_dims + 1] + list(
            range(1, self.spatial_dims + 1)
        )

    def quantize(self, inputs: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Given an input it projects it to the quantized space and returns additional tensors needed for EMA loss.

        Args:
            inputs: Encoding space tensors of shape [B, C, H, W, D].

        Returns:
            torch.Tensor: Flatten version of the input of shape [B*H*W*D, C].
            torch.Tensor: One-hot representation of the quantization indices of shape [B*H*W*D, self.num_embeddings].
            torch.Tensor: Quantization indices of shape [B,H,W,D,1]

        """
        with torch.cuda.amp.autocast(enabled=False):
            encoding_indices_view = list(inputs.shape)
            del encoding_indices_view[1]

            inputs = inputs.float()

            # Converting to channel last format
            flat_input = inputs.permute(self.flatten_permutation).contiguous().view(-1, self.embedding_dim)

            # Calculate Euclidean distances
            distances = (
                (flat_input**2).sum(dim=1, keepdim=True)
                + (self.embedding.weight.t() ** 2).sum(dim=0, keepdim=True)
                - 2 * torch.mm(flat_input, self.embedding.weight.t())
            )

            # Mapping distances to indexes
            encoding_indices = torch.max(-distances, dim=1)[1]
            encodings = torch.nn.functional.one_hot(encoding_indices, self.num_embeddings).float()

            # Quantize and reshape
            encoding_indices = encoding_indices.view(encoding_indices_view)

            return flat_input, encodings, encoding_indices

    def embed(self, embedding_indices: torch.Tensor) -> torch.Tensor:
        """
        Given encoding indices of shape [B,D,H,W,1] embeds them in the quantized space
        [B, D, H, W, self.embedding_dim] and reshapes them to [B, self.embedding_dim, D, H, W] to be fed to the
        decoder.

        Args:
            embedding_indices: Tensor in channel last format which holds indices referencing atomic
                elements from self.embedding

        Returns:
            torch.Tensor: Quantize space representation of encoding_indices in channel first format.
        """
        with torch.cuda.amp.autocast(enabled=False):
            embedding: torch.Tensor = (
                self.embedding(embedding_indices).permute(self.quantization_permutation).contiguous()
            )
            return embedding

    def distributed_synchronization(self, encodings_sum: torch.Tensor, dw: torch.Tensor) -> None:
        """
        TorchScript does not support torch.distributed.all_reduce. This function is a bypassing trick based on the
        example: https://pytorch.org/docs/stable/generated/torch.jit.unused.html#torch.jit.unused

        Args:
            encodings_sum: The summation of one hot representation of what encoding was used for each
                position.
            dw: The multiplication of the one hot representation of what encoding was used for each
                position with the flattened input.

        Returns:
            None
        """
        if self.ddp_sync and torch.distributed.is_initialized():
            torch.distributed.all_reduce(tensor=encodings_sum, op=torch.distributed.ReduceOp.SUM)
            torch.distributed.all_reduce(tensor=dw, op=torch.distributed.ReduceOp.SUM)
        else:
            pass

    def forward(self, inputs: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        flat_input, encodings, encoding_indices = self.quantize(inputs)
        quantized = self.embed(encoding_indices)

        # Use EMA to update the embedding vectors
        if self.training:
            with torch.no_grad():
                encodings_sum = encodings.sum(0)
                dw = torch.mm(encodings.t(), flat_input)

                if self.ddp_sync:
                    self.distributed_synchronization(encodings_sum, dw)

                self.ema_cluster_size.data.mul_(self.decay).add_(torch.mul(encodings_sum, 1 - self.decay))

                # Laplace smoothing of the cluster size
                n = self.ema_cluster_size.sum()
                weights = (self.ema_cluster_size + self.epsilon) / (n + self.num_embeddings * self.epsilon) * n
                self.ema_w.data.mul_(self.decay).add_(torch.mul(dw, 1 - self.decay))
                self.embedding.weight.data.copy_(self.ema_w / weights.unsqueeze(1))

        # Encoding Loss
        loss = self.commitment_cost * torch.nn.functional.mse_loss(quantized.detach(), inputs)

        # Straight Through Estimator
        quantized = inputs + (quantized - inputs).detach()

        return quantized, loss, encoding_indices


class VectorQuantizer(torch.nn.Module):
    """
    Vector Quantization wrapper that is needed as a workaround for the AMP to isolate the non fp16 compatible parts of
    the quantization in their own class.

    Args:
        quantizer (torch.nn.Module):  Quantizer module that needs to return its quantized representation, loss and index
            based quantized representation.
    """

    def __init__(self, quantizer: EMAQuantizer):
        super().__init__()

        self.quantizer: EMAQuantizer = quantizer

        self.perplexity: torch.Tensor = torch.rand(1)

    def forward(self, inputs: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        quantized, loss, encoding_indices = self.quantizer(inputs)
        # Perplexity calculations
        avg_probs = (
            torch.histc(encoding_indices.float(), bins=self.quantizer.num_embeddings, max=self.quantizer.num_embeddings)
            .float()
            .div(encoding_indices.numel())
        )

        self.perplexity = torch.exp(-torch.sum(avg_probs * torch.log(avg_probs + 1e-10)))

        return loss, quantized

    def embed(self, embedding_indices: torch.Tensor) -> torch.Tensor:
        return self.quantizer.embed(embedding_indices=embedding_indices)

    def quantize(self, encodings: torch.Tensor) -> torch.Tensor:
        output = self.quantizer(encodings)
        encoding_indices: torch.Tensor = output[2]
        return encoding_indices
